Require Import Csyntax.
Require Import Values.
Require Import Errors.
Require Import Csem.
Require Import Program.
Require Import Events.
Require Import List.
Require Import Memory.
Require Import Integers.
Require Import Globalenvs.
Require Import Maps.
Require Import Floats.
Require Cstrategy.

Ltac inj_subst H := injection H; intros; subst.

Ltac destruct_opt H :=
let v := fresh "v" in
match type of H with
| match ?E with Some x => _ | None => _ end = OK ?V => destruct E as [ v | ] _eqn; [ idtac | discriminate H ]
end.

Definition is_neutral_for_cast (ty:type) : bool :=
match ty with
| Tint I32 _ => true
| Tpointer _ => true
| Tarray _ _ => true
| Tfunction _ _ => true
| _ => false
end.

(* I wanted to do these two in one "iff" result, but that doesn't seem to work well with automation. *)
Lemma is_neutral_for_cast_sound: forall ty, is_neutral_for_cast ty = true -> Csem.neutral_for_cast ty.
intro. 
destruct ty as [ | [ | | ] | | | | | | | ];
try discriminate;
intro; constructor.
Qed.

Lemma is_neutral_for_cast_complete : forall ty, Csem.neutral_for_cast ty -> is_neutral_for_cast ty = true.
intros.
inversion H; reflexivity.
Qed.

Definition exec_cast (v:val) (ty1:type) (ty2:type) : res val  :=
match v,ty1, ty2 with
| Vint i, Tint sz1 si1, Tint sz2 si2 => OK (Vint (Csem.cast_int_int sz2 si2 i))
| Vfloat f, Tfloat sz1, Tint sz2 si2 =>
    match Csem.cast_float_int si2 f with
    | Some v => OK (Vint (Csem.cast_int_int sz2 si2 v))
    | None => Error (msg "bad cast")
    end
| Vint i, Tint sz1 si1, Tfloat sz2 => OK (Vfloat (Csem.cast_float_float sz2 (Csem.cast_int_float si1 i)))
| Vfloat f, Tfloat sz1, Tfloat sz2 => OK (Vfloat (Csem.cast_float_float sz2 f))
| Vptr _ _, _, _ => if andb (is_neutral_for_cast ty1) (is_neutral_for_cast ty2) then OK v else Error(msg "Bad ptr cast")
| Vint _, _, _ => if andb (is_neutral_for_cast ty1) (is_neutral_for_cast ty2) then OK v else Error(msg "Bad int cast")
| _, _, _ => Error(msg "bad cast")
end.

Lemma exec_cast_sound : forall v1 ty1 ty2 v2,
  exec_cast v1 ty1 ty2 = OK v2 -> Csem.cast v1 ty1 ty2 v2.
intros.
destruct v1; destruct ty1; destruct ty2; try discriminate;
cbv [exec_cast] in H;
try (rewrite Bool.andb_false_intro2 in H; solve [ discriminate | reflexivity ]);
repeat (match type of H with context c [is_neutral_for_cast ?ty] => destruct (is_neutral_for_cast ty) as [] _eqn end);
try match type of H with context c [Csem.cast_float_int ?X ?Y] => destruct (Csem.cast_float_int X Y) as [] _eqn end;
try discriminate;
simpl in H;
try (solve [inj_subst H; auto using Csem.cast, is_neutral_for_cast_sound]).
Qed.

Lemma exec_cast_complete : forall {v1 ty1 ty2 v2},
  Csem.cast v1 ty1 ty2 v2 -> exec_cast v1 ty1 ty2 = OK v2.
intros. inversion H; try reflexivity; simpl.

rewrite H0; reflexivity.

repeat rewrite is_neutral_for_cast_complete; auto; reflexivity.

inversion H0; inversion H1; auto.
Qed.

Scheme Equality for signedness.
Scheme Equality for intsize.
Scheme Equality for floatsize.
Scheme type_ind2T := Induction for type Sort Type
  with typelist_ind2T := Induction for typelist Sort Type
  with fieldlist_ind2T := Induction for fieldlist Sort Type.
(* Error: Unknown exception during scheme creation.
Scheme type_eq := Equality for type
  with typelist_eq := Equality for typelist
  with fieldlist_eq := Equality for fieldlist.
*)

Program Fixpoint type_eq (ty1 ty2 : type) : {ty1 = ty2} + {ty1 <> ty2} :=
match ty1, ty2 with
| Tvoid, Tvoid => left _
| Tint sz1 sg1, Tint sz2 sg2 =>
    if intsize_eq_dec sz1 sz2 then
      if signedness_eq_dec sg1 sg2 then left _ else right _
    else right _
| Tfloat sz1, Tfloat sz2 =>
    if floatsize_eq_dec sz1 sz2 then left _ else right _
| Tpointer ty1', Tpointer ty2' =>
    if type_eq ty1' ty2' then left _ else right _
| Tarray ty1' n1, Tarray ty2' n2 =>
    if type_eq ty1' ty2' then
      if ZArith_dec.Z_eq_dec n1 n2 then left _ else right _
    else right _
| Tfunction args1 ty1', Tfunction args2 ty2' =>
    if typelist_eq args1 args2 then
      if type_eq ty1' ty2' then left _ else right _
    else right _
| Tstruct id1 flds1, Tstruct id2 flds2 =>
    if AST.ident_eq id1 id2 then
      if fieldlist_eq flds1 flds2 then left _ else right _
    else right _
| Tunion id1 flds1, Tunion id2 flds2 =>
    if AST.ident_eq id1 id2 then
      if fieldlist_eq flds1 flds2 then left _ else right _
    else right _
| Tcomp_ptr id1, Tcomp_ptr id2 =>
    if AST.ident_eq id1 id2 then left _ else right _
| _, _ => right _
end
with typelist_eq (l1 l2 : typelist) : {l1 = l2} + {l1 <> l2} :=
match l1, l2 with
| Tnil, Tnil => left _
| Tcons h1 t1, Tcons h2 t2 =>
    if type_eq h1 h2 then
      if typelist_eq t1 t2 then left _ else right _
    else right _
| _,_ => right _
end
with fieldlist_eq (l1 l2 : fieldlist) : {l1 = l2} + {l1 <> l2} :=
match l1, l2 with
| Fnil, Fnil => left _
| Fcons id1 ty1 tl1, Fcons id2 ty2 tl2 =>
    if AST.ident_eq id1 id2 then
      if type_eq ty1 ty2 then
        if fieldlist_eq tl1 tl2 then left _ else right _
      else right _
    else right _
| _,_ => right _
end.
Next Obligation of typelist_eq.
destruct l1; destruct l2; intuition. injection H1. intros; subst. apply (H t0 l2 t0 l2). auto.
Qed.
Next Obligation of type_eq.
intro E; subst; destruct ty2;
[ eapply H7 | eapply H | eapply H0 | eapply H1 | eapply H2 | eapply H3 | eapply H4 | eapply H5 | eapply H6 ];
auto.
Qed.
Next Obligation of fieldlist_eq.
intro E; subst; destruct l2; auto. eapply H; auto.
Qed.
Solve All Obligations using program_simpl; intuition discriminate.

Lemma dec_eq_true : forall {T:Type}, forall f : (forall x y : T, {x = y} + {x <> y}),
  forall x:T, exists p, f x x = left p.
intros. destruct (f x x). exists e; reflexivity.
destruct n. reflexivity.
Qed.

Lemma dec_eq_false : forall {T:Type}, forall f : (forall x y : T, {x = y} + {x <> y}),
  forall {x y}, x <> y -> exists p, f x y = right p.
intros. destruct (f x y). contradiction.
exists n. reflexivity.
Qed.


Section LRED.

Open Local Scope error_monad_scope.

Hint Constructors Csem.lred.

Variable ge:Csem.genv.
Variable en:Csem.env.

Definition exec_lred (e:expr) (m:mem) : res (expr * mem) :=
match e with
| Evar x ty =>
    match en!x with
    | Some (b, ty') => match type_eq ty ty' with
                                 | left _ => OK (Eloc b Int.zero ty, m)
                                 | right _ => Error (msg "type mismatch")
                                 end
    | None =>
        match Genv.find_symbol ge x with
        | Some b =>
            match Csem.type_of_global ge b with
            | Some ty' => match type_eq ty ty' with
                                   | left _ => OK (Eloc b Int.zero ty, m)
                                 | right _ => Error (msg "type mismatch")
                                  end
            | None => Error (msg "type_of_global failed")
            end
        | None => Error (msg "missing symbol")
        end
    end
| Ederef (Eval (Vptr b ofs) ty1) ty =>
    OK (Eloc b ofs ty, m)
| Efield (Eloc b ofs (Tstruct id fList)) f ty =>
    do delta <- field_offset f fList;
    OK (Eloc b (Int.add ofs (Int.repr delta)) ty, m)
| Efield (Eloc b ofs (Tunion id fList)) f ty =>
    OK (Eloc b ofs ty, m)
| _ => Error (msg "bad lred")
end.

Lemma exec_lred_sound : forall e m e' m',
  exec_lred e m = OK (e', m') -> Csem.lred ge en e m e' m'.
Proof with try discriminate.
intros; destruct e; try discriminate; simpl in H.

destruct (en!x) as [[b ty']| ] _eqn.
  destruct (type_eq ty ty'); try discriminate; inj_subst H; auto.

  destruct (Genv.find_symbol ge x) as [] _eqn...
  destruct (Csem.type_of_global ge b) as [] _eqn...
  destruct (type_eq ty t); try discriminate; inj_subst H; auto.

destruct e...
destruct ty0...
  monadInv H. auto.

  inj_subst H; auto.

destruct e...
destruct v...
simplify_eq H.
inj_subst H; auto.
Qed.

Lemma exec_lred_complete : forall {e m e' m'},
  Csem.lred ge en e m e' m' -> exec_lred e m = OK (e', m').
intros; inversion H; subst; simpl; auto.
  rewrite H0; simpl. destruct (dec_eq_true type_eq ty). rewrite H1; reflexivity.

  rewrite H0; rewrite H1; simpl. rewrite H2; simpl; destruct (dec_eq_true type_eq ty). rewrite H3; reflexivity.

  rewrite H0; reflexivity.
Qed.

End LRED.


Definition exec_bool_val (v:val) (ty:type) : res bool :=
match v,ty with
| Vint i   , Tint _ _   => OK (negb (Int.eq i Int.zero))
| Vint i   , Tpointer _ => OK (negb (Int.eq i Int.zero))
| Vptr _ _ , Tint _ _   => OK true
| Vptr _ _ , Tpointer _ => OK true
| Vfloat f , Tfloat _   => OK (negb (Float.cmp Ceq f Float.zero))
| _, _ => Error (msg "non-boolean value")
end.

Lemma eq_true_true : forall {x y}, Int.eq x y = true -> x = y.
intros. generalize (Int.eq_spec x y). rewrite H; auto.
Qed.

Lemma eq_false_false : forall {x y}, Int.eq x y = false -> x <> y.
intros. generalize (Int.eq_spec x y). rewrite H; auto.
Qed.

Lemma exec_bool_val_sound_true : forall {v ty}, exec_bool_val v ty = OK true -> Csem.is_true v ty.
intros; destruct v; destruct ty; simpl in H; try discriminate; try inj_subst H; auto using Csem.is_true.
  apply Csem.is_true_int_int. apply eq_false_false. symmetry in H0. apply Bool.negb_sym in H0. apply H0.

  apply Csem.is_true_int_pointer. apply eq_false_false. symmetry in H0. apply Bool.negb_sym in H0. apply H0.

  apply Csem.is_true_float. symmetry in H0. apply Bool.negb_sym in H0. apply H0.
Qed.

Lemma exec_bool_val_sound_false : forall {v ty}, exec_bool_val v ty = OK false -> Csem.is_false v ty.
intros; destruct v; destruct ty; simpl in H; try discriminate; try inj_subst H; auto using Csem.is_false.
  symmetry in H0. apply Bool.negb_sym in H0. rewrite (eq_true_true H0).
  apply Csem.is_false_int.

  symmetry in H0. apply Bool.negb_sym in H0. rewrite (eq_true_true H0).
  apply Csem.is_false_pointer.

  apply Csem.is_false_float. symmetry in H0. apply Bool.negb_sym in H0. apply H0.
Qed.

Lemma exec_bool_val_true : forall {v ty}, Csem.is_true v ty -> exec_bool_val v ty = OK true.
intros; inversion H; subst; simpl; auto.
  rewrite Int.eq_false; auto.
  rewrite Int.eq_false; auto.
  rewrite H0; auto.
Qed.

Lemma exec_bool_val_false : forall {v ty}, Csem.is_false v ty -> exec_bool_val v ty = OK false.
intros; inversion H; subst; simpl; auto.
  rewrite H0; auto.
Qed.


Definition assert_type (ty ty':type) : res (ty = ty') :=
match type_eq ty ty' with
| left p => OK p
| right _ => Error (msg "type mismatch")
end.

Lemma assert_type_true : forall ty, exists p, assert_type ty ty = OK p.
intro. unfold assert_type. destruct (dec_eq_true type_eq ty). exists x; rewrite H; reflexivity.
Qed.

Section RRED.

Hint Constructors Csem.rred.

Variable ge:Csem.genv.
Variable en:Csem.env.

Open Local Scope error_monad_scope.

Definition exec_rred (e:expr) (m:mem) : res (expr * mem) :=
match e with
| Evalof (Eloc b ofs ty) ty' =>
    do p <- assert_type ty ty';
    match Csem.load_value_of_type ty m b ofs with
    | Some v => OK (Eval v ty, m)
    | None => Error (msg "Load failed")
    end
| Eaddrof (Eloc b ofs ty1) ty => OK (Eval (Vptr b ofs) ty, m)
| Eunop op (Eval v1 ty1) ty =>
    match Csem.sem_unary_operation op v1 ty1 with
    | Some v => OK (Eval v ty, m)
    | None => Error (msg "unop failed")
    end
| Ebinop op (Eval v1 ty1) (Eval v2 ty2) ty =>
    match Csem.sem_binary_operation op v1 ty1 v2 ty2 m with
    | Some v => OK (Eval v ty, m)
    | None => Error (msg "binop failed")
    end
| Ecast (Eval v1 ty1) ty =>
    do v <- exec_cast v1 ty1 ty;
    OK (Eval v ty, m)
| Econdition (Eval v1 ty1) r1 r2 ty =>
    do b <- exec_bool_val v1 ty1;
    let r := if b then r1 else r2 in
    match type_eq (typeof r) ty with
    | left _ => OK (Eparen r ty, m)
    | right _ => Error (msg "type mismatch in Econdition")
    end
| Esizeof ty1 ty => OK (Eval (Vint (Int.repr (sizeof ty1))) ty, m)
| Eassign (Eloc b ofs ty1) (Eval v2 ty2) ty1' =>
    do p <- assert_type ty1 ty1';
    do v <- exec_cast v2 ty2 ty1;
    match Csem.store_value_of_type ty1 m b ofs v with
    | Some m' => OK (Eval v ty1, m')
    | None => Error (msg "bad store in assignment")
    end
| Eassignop op (Eloc b ofs ty1) (Eval v2 ty2) tyres ty1' =>
    match Csem.load_value_of_type ty1 m b ofs with
    | None => Error (msg "load failed in assignop")
    | Some v1 =>
        do p <- assert_type ty1 ty1';
        match Csem.sem_binary_operation op v1 ty1 v2 ty2 m with
        | None => Error (msg "bad op in assignop")
        | Some v =>
            do v' <- exec_cast v tyres ty1;
            match Csem.store_value_of_type ty1 m b ofs v' with
            | Some m' => OK (Eval v' ty1, m')
            | None => Error (msg "bad store in assignop")
            end
        end
    end
| Epostincr id' (Eloc b ofs ty) ty' =>
    do p <- assert_type ty ty';
    match Csem.load_value_of_type ty m b ofs with
    | None => Error (msg "load failed in postincr")
    | Some v1 =>
        match Csem.sem_incrdecr id' v1 ty with
        | None => Error (msg "bad op in postincr")
        | Some v2 =>
            do v3 <- exec_cast v2 (typeconv ty) ty;
            match Csem.store_value_of_type ty m b ofs v3 with
            | Some m' => OK (Eval v1 ty, m')
            | None => Error (msg "bad store in postincr")
            end
        end
    end
| Ecomma (Eval v ty1) r2 ty =>
    do p <- assert_type (typeof r2) ty;
    OK (r2, m)
| Eparen r ty =>
    do p <- assert_type (typeof r) ty;
    OK (r, m)
| _ => Error (msg "undefined rred")
end.

Lemma exec_rred_sound : forall e m e' m',
 exec_rred e m = OK (e',m') ->
 Csem.rred e m e' m'.
Proof with try discriminate.
intros; destruct e; try discriminate; simpl in H.
  destruct e... monadInv H. destruct_opt EQ0. injection EQ0; intros; subst; auto.

  destruct e... inj_subst H; auto.

  destruct e... destruct_opt H. inj_subst H; auto.

  destruct e1... destruct e2... destruct_opt H. inj_subst H;auto.

  destruct e... monadInv H. auto using exec_cast_sound. 

  destruct e1... monadInv H.
  match type of EQ0 with (if ?E then _ else _) = _ => destruct E as [] _eqn end...
  inj_subst EQ0; destruct x; [ apply exec_bool_val_sound_true in EQ | apply exec_bool_val_sound_false in EQ ];
  auto.

  inj_subst H; auto.

  destruct e1... destruct e2... monadInv H. destruct_opt EQ2. inj_subst EQ2. auto using exec_cast_sound.

  destruct e1... destruct e2... destruct_opt H. monadInv H. destruct_opt EQ0. monadInv EQ0.
  destruct_opt EQ2. inj_subst EQ2. eauto using exec_cast_sound.

  destruct e... monadInv H. repeat destruct_opt EQ0. monadInv EQ0. destruct_opt EQ2.
  inj_subst EQ2; eauto using exec_cast_sound.

  destruct e1... monadInv H. auto.

  monadInv H. auto.
Qed.

Lemma exec_rred_complete : forall {e m e' m'},
 Csem.rred e m e' m' ->
 exec_rred e m = OK (e',m').
intros; inversion H; subst; simpl;
try match goal with |- context c [ assert_type ?ty ?ty ] => destruct (assert_type_true ty) as [ p E ]; rewrite E end;
simpl; subst; try rewrite H0; simpl; auto.

rewrite (exec_cast_complete H0). auto.

rewrite (exec_bool_val_true H0). simpl. destruct (dec_eq_true type_eq (typeof r1)). rewrite H1; auto.

rewrite (exec_bool_val_false H0). simpl. destruct (dec_eq_true type_eq (typeof r2)). rewrite H1; auto.

rewrite (exec_cast_complete H0); simpl; rewrite H1; auto.

rewrite H1; simpl; rewrite (exec_cast_complete H2); simpl; rewrite H3; auto.

rewrite H1; rewrite (exec_cast_complete H2); simpl; rewrite H3; auto.
Qed.

End RRED.

Section CALLS.

Hint Constructors Csem.cast_arguments.
Hint Constructors Csem.callred.

Variable ge:Csem.genv.
Variable en:Csem.env.

Open Local Scope error_monad_scope.

Fixpoint exec_cast_args (el:exprlist) (tl:typelist) : res (list val) :=
match el,tl with
| Enil, Tnil => OK nil
| Econs e etl, Tcons targ1 ttl =>
    match e with
    | Eval v ty =>
        do v1 <- exec_cast v ty targ1;
        do vtl <- exec_cast_args etl ttl;
        OK (cons v1 vtl)
    | _ => Error (msg "argument not reduced")
    end
| _,_ => Error (msg "function has wrong number of arguments")
end.

Lemma exec_cast_args_sound : forall el tl vl,
  exec_cast_args el tl = OK vl -> Csem.cast_arguments el tl vl.
intro el; induction el; intros; destruct tl; simpl in H; try discriminate.
  inj_subst H; auto.

  destruct r1; try discriminate.
  monadInv H. auto using exec_cast_sound.
Qed.

Lemma exec_cast_args_complete : forall {el tl vl},
  Csem.cast_arguments el tl vl -> exec_cast_args el tl = OK vl.
intros. induction H; auto.

simpl. rewrite (exec_cast_complete H). simpl. rewrite IHcast_arguments. auto.
Qed.

Definition exec_callred (e:expr) : res (fundef * list val * type) :=
match e with
| Ecall (Eval vf tyf) el ty =>
    match classify_fun tyf with
    | fun_case_f tyargs tyres =>
        match Genv.find_funct ge vf with
        | Some fd =>
            match type_eq (type_of_fundef fd) (Tfunction tyargs tyres) with
            | left _ => do vargs <- exec_cast_args el tyargs; OK (fd,vargs,ty)
            | right _ => Error (msg "function type mismatch")
            end
        | None => Error (msg "missing function")
        end
    | fun_case_default => Error (msg "function value hasn't a function type")
    end
| _ => Error (msg "callred used on expression that isn't a call on a reduced value")
end.

Lemma exec_callred_sound : forall e fd vargs ty,
  exec_callred e = OK (fd,vargs,ty) -> Csem.callred ge e fd vargs ty.
Proof with try discriminate.
intros. destruct e... simpl in H.
  destruct e... destruct (classify_fun ty1) as [tyargs tyres | ] _eqn... destruct_opt H.
  destruct (type_eq (type_of_fundef v0) (Tfunction tyargs tyres))...
  monadInv H. eauto using exec_cast_args_sound.
Qed.

Lemma exec_callred_complete : forall {e fd vargs ty},
  Csem.callred ge e fd vargs ty -> exec_callred e = OK (fd,vargs,ty).
intros. inversion H; subst; auto.
simpl. rewrite H0, H2, H3. destruct (dec_eq_true type_eq (Tfunction tyargs tyres)).
rewrite H4. rewrite (exec_cast_args_complete H1). reflexivity.
Qed.

End CALLS.


Section NOT_STUCK.

(* The inductive non-deterministic semantics requires that there are no stuck
   subexpressions, even if they're not the next subexpression to be reduced.
   This ensures that expressions which should go wrong aren't masked by some
   other subexpression that reduces forever (e.g., the f() + 10/0 example from
   Csem.v).

   The not_stuck definition is quantified over all subcontexts.  Here we give
   a recursive function which does the check in a single pass. *)

Variable ge:genv.
Variable en:env.

(* We distinguish between expressions which are not stuck because they are
   reduced, and those which have a reducible subexpression. *)
Inductive stuckness : Type :=
| Reduced : stuckness
| Reducible : stuckness
| Stuck : errmsg -> kind -> expr -> stuckness.

(* I plan to clean this up shortly. *)
Fixpoint check_not_stuck (k:kind) (e:expr) (m:mem) {struct e} : stuckness :=
match k,e with
| RV, Eval _ _
| LV, Eloc _ _ _ => Reduced

(* lvalues *)
| LV, Evar _ _ =>
    match exec_lred ge en e m with OK _  => Reducible | Error msg => Stuck msg LV e end
| LV, Ederef e' _ =>
    match check_not_stuck RV e' m with
    | Reduced => match exec_lred ge en e m with OK _  => Reducible | Error msg => Stuck msg LV e end
    | Reducible => Reducible
    | Stuck msg k e => Stuck msg k e
    end
| LV, Efield e' _ _ =>
    match check_not_stuck LV e' m with
    | Reduced => match exec_lred ge en e m with OK _ => Reducible | Error msg => Stuck msg LV e end
    | Reducible => Reducible
    | Stuck msg k e => Stuck msg k e
    end

(* poor, lonely sizeof is the only rvalue with no subexpressions. *)
| RV, Esizeof _ _ =>
    match exec_rred e m with OK _ => Reducible | Error msg => Stuck msg RV e end

(* rvalues with single reducible lvalue position *)
| RV, Evalof e' _
| RV, Eaddrof e' _
| RV, Epostincr _ e' _ =>
    match check_not_stuck LV e' m with
    | Reduced => match exec_rred e m with OK _ => Reducible | Error msg => Stuck msg RV e end
    | Reducible => Reducible
    | Stuck msg k e => Stuck msg k e
    end

(* rvalues with single reducible rvalue position *)
| RV, Eunop _ e' _
| RV, Ecast e' _
| RV, Econdition e' _ _ _
| RV, Ecomma e' _ _
| RV, Eparen e' _ =>
    match check_not_stuck RV e' m with
    | Reduced => match exec_rred e m with OK _ => Reducible | Error msg => Stuck msg RV e end
    | Reducible => Reducible
    | Stuck msg k e => Stuck msg k e
    end

(* two rvalue positions *)
| RV, Ebinop _ e1 e2 _ =>
    match check_not_stuck RV e1 m with
    | Reduced =>
        match check_not_stuck RV e2 m with
        | Reduced => match exec_rred e m with OK _ => Reducible | Error msg => Stuck msg RV e end
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Reducible =>
        match check_not_stuck RV e2 m with
        | Reduced
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Stuck msg k e => Stuck msg k e
    end

(* assignments have an lvalue and an rvalue position *)
| RV, Eassign e1 e2 _
| RV, Eassignop _ e1 e2 _ _ =>
    match check_not_stuck LV e1 m with
    | Reduced =>
        match check_not_stuck RV e2 m with
        | Reduced => match exec_rred e m with OK _ => Reducible | Error msg => Stuck msg RV e end
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Reducible =>
        match check_not_stuck RV e2 m with
        | Reduced
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Stuck msg k e => Stuck msg k e
    end

(* function calls have an argument list as well as an rvalue *)
| RV, Ecall ef eargs _ =>
    match check_not_stuck RV ef m with
    | Reduced =>
        match check_list_not_stuck eargs m with
        | Reduced => match exec_callred ge e with OK _ => Reducible | Error msg => Stuck msg RV e end
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Reducible =>
        match check_list_not_stuck eargs m with
        | Reduced
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Stuck msg k e => Stuck msg k e
    end

(* everything else is nonsense *)
| _, _ => Stuck (msg "Bad kind,expr pair given to check_not_stuck") k e
end
with check_list_not_stuck (el:exprlist) (m:mem) {struct el} : stuckness :=
match el with
| Enil => Reduced
| Econs e el =>
    match check_not_stuck RV e m with
    | Reduced => check_list_not_stuck el m
    | Reducible =>
        match check_list_not_stuck el m with
        | Reduced
        | Reducible => Reducible
        | Stuck msg k e => Stuck msg k e
        end
    | Stuck msg k e => Stuck msg k e
    end
end.

Functional Scheme check_not_stuck_ind := Induction for check_not_stuck Sort Prop
with check_list_not_stuck_ind := Induction for check_list_not_stuck Sort Prop.

(* We split not_imm_stuck into reduced and reducible versions. *)

Inductive not_imm_stuck_reduced: kind -> expr -> mem -> Prop :=
  | not_stuck_val: forall v ty m,
      not_imm_stuck_reduced RV (Eval v ty) m
  | not_stuck_loc: forall b ofs ty m,
      not_imm_stuck_reduced LV (Eloc b ofs ty) m.

Inductive not_imm_stuck_reducible: kind -> expr -> mem -> Prop :=
  | not_stuck_lred: forall to C e m e' m',
      lred ge en e m e' m' ->
      context LV to C ->
      not_imm_stuck_reducible to (C e) m
  | not_stuck_rred: forall to C e m e' m',
      rred e m e' m' ->
      context RV to C ->
      not_imm_stuck_reducible to (C e) m
  | not_stuck_callred: forall to C e m fd args ty,
      callred ge e fd args ty ->
      context RV to C ->
      not_imm_stuck_reducible to (C e) m.

Lemma not_imm_stuck_inc1 : forall k e m,
  not_imm_stuck_reduced k e m -> not_imm_stuck ge en k e m.
intros; inversion H; eauto using not_imm_stuck.
Qed.

Lemma not_imm_stuck_inc2 : forall k e m,
  not_imm_stuck_reducible k e m -> not_imm_stuck ge en k e m.
intros; inversion H; eauto using not_imm_stuck.
Qed.

(* The idea here is that not only is the expression (and contexts) not stuck,
   but there is a reduction available. *)
Definition not_stuck_reducible (k':kind) (e:expr) (m:mem) : Prop :=
  not_imm_stuck_reducible k' e m /\
  forall k C e',
  context k k' C -> e = C e' -> not_imm_stuck ge en k e' m.

Definition list_reduced (el:exprlist) (m:mem) : Prop :=
  forall k C e,
    contextlist k C -> el = C e -> not_imm_stuck_reduced k e m.

Definition list_reducible (el:exprlist) (m:mem) : Prop :=
  exists k C e,
    contextlist k C /\ el = C e /\ not_imm_stuck_reducible k e m.

Definition list_not_stuck (el:exprlist) (m:mem) : Prop :=
  forall k C e,
    contextlist k C -> el = C e -> not_imm_stuck ge en k e m.

Hint Constructors context.
Hint Resolve Cstrategy.context_compose.

Lemma not_imm_stuck_step : forall k k' C e m,
  not_imm_stuck_reducible k e m ->
  context k k' C ->
  not_imm_stuck_reducible k' (C e) m.
intros. inversion H; subst.
  eapply (not_stuck_lred _ (fun x => (C (C0 x)))); eauto.
  eapply (not_stuck_rred _ (fun x => (C (C0 x)))); eauto.
  eapply (not_stuck_callred _ (fun x => (C (C0 x)))); eauto.
Qed.

(* Break up tuples in results to allow reduction. *)
Ltac break_optpair :=
  repeat
    lazymatch goal with
    | H: @eq (res (prod (prod _ _) _)) _ (OK ((_,_),_)) |- _ => fail
    | H: @eq (res (prod (prod _ _) _)) _ (OK (?X,_)) |- _ => destruct X
    | H: @eq (res (prod _ _)) _ (OK (_,_)) |- _ => fail
    | H: @eq (res (prod _ _)) _ (OK ?X) |- _ => destruct X end.

(* Extract the context and apply some tactic. *)
Ltac applyctx t :=
let E :=
  match goal with
  | IMM : not_imm_stuck_reducible _ ?E _ |- _ => E
  | IMM : not_imm_stuck_reduced _ ?E _ |- _ => E
  end in
  match goal with
  | |- not_imm_stuck_reducible _ (?CS E ?E2) _ => t (fun x => CS x E2)
  | |- not_imm_stuck_reducible _ (?CS E ?E2 ?E3) _ => t (fun x => CS x E2 E3)
  | |- not_imm_stuck_reducible _ (?CS ?E1 E ?E3) _ => t (fun x => CS E1 x E3)
  | |- not_imm_stuck_reducible _ (?CS E ?E2 ?E3 ?E4) _ => t (fun x => CS x E2 E3 E4)
  | |- not_imm_stuck_reducible _ (?CS ?E1 E ?E3 ?E4) _ => t (fun x => CS E1 x E3 E4)
  | |- not_imm_stuck_reducible _ (?CS ?E1 ?E2 E ?E4) _ => t (fun x => CS E1 E2 x E4)
  | |- not_imm_stuck_reducible _ (?CS E ?E2 ?E3 ?E4 ?E5) _ => t (fun x => CS x E2 E3 E4 E5)
  | |- not_imm_stuck_reducible _ (?CS ?E1 E ?E3 ?E4 ?E5) _ => t (fun x => CS E1 x E3 E4 E5)
  | |- not_imm_stuck_reducible _ (?CS ?E1 ?E2 E ?E4 ?E5) _ => t (fun x => CS E1 E2 x E4 E5)
  | |- not_imm_stuck_reducible _ (?CS ?E1 ?E2 ?E3 E ?E5) _ => t (fun x => CS E1 E2 E3 x E5)
  end
.

Ltac stepctx _ :=
 let stept := fun ctx => (eapply (not_imm_stuck_step _ _ ctx)); eauto in
 applyctx stept.

Ltac normal_reducible steptac SUB SUB0 H :=
  (* prove not_stuck_reducible, which involves using steptac to show not_imm_stuck_reducible,
     show that subcontexts of reduced expressions are the whole already-reduced expression,
     and applying SUB for all other subcontexts. *)
  split;
  [ steptac 0
  | intros k0 C e1 CTX ECTX; inversion CTX; subst; subst; try discriminate;
    solve
    [ apply not_imm_stuck_inc2; steptac 0
    | apply not_imm_stuck_inc1; injection ECTX; intros; subst; inversion H; subst; auto; inversion R
    | apply not_imm_stuck_inc1; injection ECTX; intros; subst; inversion H; subst; auto; inversion R0
    | injection ECTX; intros; subst; apply (SUB _ _ _ H (eq_refl _))
    | injection ECTX; intros; subst; apply (SUB0 _ _ _ H (eq_refl _))
    ]
  ].

Ltac reduce_term _ :=
match goal with
| E:exec_lred _ _ ?e _ = _ |- _ => eapply (not_stuck_lred _ (fun x => x) e); auto; apply exec_lred_sound; apply E
| E:exec_rred ?e _ = _ |- _ => eapply (not_stuck_rred _ (fun x => x) e); auto; apply exec_rred_sound; apply E
| E:exec_callred _ ?e _ = _ |- _ => eapply (not_stuck_callred _ (fun x => x) e); auto; apply exec_callred_sound; apply E
end.

Lemma check_not_stuck_sound : forall k e m,
  match check_not_stuck k e m with
  | Reduced => not_imm_stuck_reduced k e m
  | Reducible => not_stuck_reducible k e m
  | Stuck _ _ _ => True 
  end.
intros; apply check_not_stuck_ind with (P :=
  fun k e m s => match s with
                  | Reduced => not_imm_stuck_reduced k e m
                  | Reducible => not_stuck_reducible k e m
                  | Stuck _ _ _ => True
                  end)
  (P0 := fun el m s =>
   match s with
   | Reduced => list_reduced el m
   | Reducible => list_reducible el m /\ list_not_stuck el m
   | Stuck _ _ _ => True
   end);
intros; subst; auto using not_imm_stuck_reduced; try intros k0 C e0 H1 H2; break_optpair;
(* Tidy up facts from recursive calls.*)
repeat
match goal with X:match ?E with Reduced => _ | Reducible => _ | Stuck _ _ _ => _ end,
                Y: ?E = _
                |- _ => let H := fresh "R" in rewrite Y in X; rename X into H; clear Y end;
repeat
match goal with R:not_stuck_reducible ?k ?e ?m |- _ => let IMM := fresh "IMM" in let SUB := fresh "SUB" in destruct R as [IMM SUB] end;
try normal_reducible stepctx SUB SUB0 H;
try normal_reducible reduce_term SUB SUB0 H.

(* We do the function cases by hand. *)
split.
  eapply (not_stuck_callred _ (fun x => x) (Ecall ef eargs _x)); auto; apply exec_callred_sound; apply e5.
  intros k0 C e1 CTX ECTX; inversion CTX; subst; subst; try discriminate.
    apply not_imm_stuck_inc2. eapply (not_stuck_callred _ (fun x => x) (Ecall ef eargs _x)); auto; apply exec_callred_sound; apply e5.
    apply not_imm_stuck_inc1; injection ECTX; intros; subst; inversion H; subst; auto; inversion R0.
    apply not_imm_stuck_inc1; injection ECTX; intros; subst. apply (R k0 C0 e1); auto.

destruct R as [[k1 [C1 [e1 [CTX1 [ECTX1 R1]]]]] NS]. subst. split.
  apply not_imm_stuck_step with (k:=k1) (C:=(fun x => Ecall ef (C1 x) _x)); auto.
  intros k0 C e0 CTX ECTX; inversion CTX; subst; subst; try discriminate.
    apply not_imm_stuck_inc2. apply not_imm_stuck_step with (k:=k1) (C:=(fun x => Ecall ef (C1 x) _x)); auto.
    apply not_imm_stuck_inc1; injection ECTX; intros; subst; inversion H; subst; auto; inversion R0.
    injection ECTX; intros; subst. rewrite H1 in NS. apply NS with (C:=C0);auto.

split.
  stepctx 0.
  intros k0 C e1 CTX ECTX; inversion CTX; subst; subst; try discriminate.
    apply not_imm_stuck_inc2. stepctx 0.
    injection ECTX; intros; subst; apply (SUB _ _ _ H (eq_refl _)).
    apply not_imm_stuck_inc1; injection ECTX; intros; subst. apply (R k0 C0 e1); auto.

destruct R as [[k1 [C1 [e1 [CTX1 [ECTX1 R1]]]]] NS]. subst. split.
  apply not_imm_stuck_step with (k:=k1) (C:=(fun x => Ecall ef (C1 x) _x)); auto.
  intros k0 C e0 CTX ECTX; inversion CTX; subst; subst; try discriminate.
    apply not_imm_stuck_inc2. apply not_imm_stuck_step with (k:=k1) (C:=(fun x => Ecall ef (C1 x) _x)); auto.
    injection ECTX; intros; subst; apply (SUB _ _ _ H (eq_refl _)).
    injection ECTX; intros; subst. rewrite H1 in NS. apply NS with (C:=C0);auto.

inversion H1; intros; subst; inversion H2.

destruct (check_list_not_stuck el0 m0) as [] _eqn; auto.
  intros k1 C e1 CTX ECTX. inversion CTX; subst.
    injection ECTX; intros; subst. inversion H; intros; subst; inversion R; intros; subst; auto.
    injection ECTX; intros; subst. apply H0 with (C:=C0); auto.

destruct H0 as [[k1 [C1 [e1 [CTX1 [ECTX1 R1]]]]] NS]. split.
  exists k1. exists (fun x => Econs e0 (C1 x)). exists e1. intuition. auto using ctx_list_tail. congruence.
  intros k2 C e2 CTX ECTX. inversion CTX; subst.
    apply not_imm_stuck_inc1. injection ECTX; intros; subst. inversion H; intros; subst; inversion R; intros; subst; auto.
    injection ECTX; intros; subst. apply NS with (C:=C0); auto.

split.
  exists RV. exists (fun x => Econs x el0). exists e0. intuition. auto using ctx_list_head.
  intros k2 C e2 CTX ECTX. inversion CTX; subst.
    injection ECTX; intros; subst. apply SUB with (C:=C0);auto.
    injection ECTX; intros; subst. apply not_imm_stuck_inc1. apply R with (C:=C0); auto.

destruct R as [[k1 [C1 [e1 [CTX1 [ECTX1 R1]]]]] NS]. subst. split.
  exists RV. exists (fun x => Econs x (C1 e1)). exists e0. intuition. auto using ctx_list_head.
  intros k2 C e2 CTX ECTX. inversion CTX; subst.
    injection ECTX; intros; subst. apply SUB with (C:=C0);auto.
    injection ECTX; intros; subst. apply NS with (C:=C0); auto.
Qed.

(* We proved this above, but don't have a combined induction principle to
   keep the result, so let's recreate it. *)
Lemma check_list_not_stuck_sound : forall el m,
   match check_list_not_stuck el m with
   | Reduced => list_reduced el m
   | Reducible => list_reducible el m /\ list_not_stuck el m
   | Stuck _ _ _ => True
   end.
intros.
induction el; simpl.
  intros k C e CTX ECTX. inversion CTX; subst; discriminate.

assert (Shead := check_not_stuck_sound RV r1 m).
destruct (check_not_stuck RV r1 m); auto.
  destruct (check_list_not_stuck el m); auto.
    intros k C e CTX ECTX. inversion CTX; subst; try discriminate.
      inversion H; subst; try discriminate; injection ECTX; intros; subst; try discriminate; auto; inversion Shead.
      injection ECTX; intros; subst. apply (IHel k C0 e); auto.
    destruct IHel as [ [k1 [C1 [e1 [CTX [ECTX NSIM]]]]] NS ].
    split.
      exists k1 (fun x => Econs r1 (C1 x)) e1; intuition. auto using contextlist. congruence.
      clear k1 C1 e1 CTX ECTX NSIM. intros k2 C2 e2 CTX2 ECTX2. inversion CTX2; subst.
        inversion H; subst; try discriminate; injection ECTX2; intros; subst; try discriminate; auto; inversion Shead; subst; auto using not_imm_stuck_inc1.
        injection ECTX2; intros; subst. apply (NS k2 C e2); auto.

  destruct (check_list_not_stuck el m); auto.
  split.
    exists RV (fun x => Econs x el) r1. intuition auto using contextlist. apply Shead.
    intros k1 C1 e1 CTX1 ECTX1. inversion CTX1; subst; injection ECTX1; intros; subst.
    apply ((proj2 Shead) k1 C e1); auto.
    apply not_imm_stuck_inc1. apply (IHel k1 C e1); auto.
      
  split.
    exists RV (fun x => Econs x el) r1. intuition auto using contextlist. apply Shead.
    intros k1 C1 e1 CTX1 ECTX1. inversion CTX1; subst; injection ECTX1; intros; subst.
    apply ((proj2 Shead) k1 C e1); auto.
    apply ((proj2 IHel) k1 C e1); auto.
Qed.

(* For the completeness result, we first need to generalise not_stuck for any
   kind. *)

Definition not_stuck_for (k':kind) (e:expr) (m:mem) : Prop :=
  forall k C e',
  context k k' C -> e = C e' -> not_imm_stuck ge en k e' m.

(* Extract not_imm_stuck for the current expression. *)
Ltac top_not_stuck H :=
  match type of H with not_stuck_for ?k ?e _ =>
    assert (H' := H k (fun x => x) e (ctx_top _) (eq_refl _))
  end.

(* Inversion of current expression's not_imm_stuckness with vaguely sane
   identifiers. *)
Ltac inv_imm_stuck H :=
  top_not_stuck H;
  inversion H' as [ | | to C e99 m99 e' m' RED CTX E1 E2 E3
                      | to C e99 m99 e' m' RED CTX E1 E2 E3
                      | to C e99 m0 fd args ty99 RED CTX E1 E2 E3 ]; subst.

(* Apply a predicate to all of the subexpressions in a subcontext. *)
Fixpoint req_subexprs (e:expr) (P:kind -> expr -> Prop) : Prop :=
  match e with
  | Evar _ _
  | Esizeof _ _
    => True

  | Efield e _ _
  | Evalof e _
  | Eaddrof e _
  | Epostincr _ e _
    => P LV e

  | Ederef e _
  | Eunop _ e _
  | Ecast e _
  | Econdition e _ _ _
  | Ecomma e _ _
  | Eparen e _
    => P RV e

  | Ebinop _ e1 e2 _
    => P RV e1 /\ P RV e2
  | Eassign e1 e2 _
  | Eassignop _ e1 e2 _ _
    => P LV e1 /\ P RV e2

  | Ecall e1 eargs _ => P RV e1 /\ req_subexprs_list eargs P

    (* we're interested in reductions, so no value expressions *)
  | Eval _ _
  | Eloc _ _ _ => False
  end
with req_subexprs_list (el:exprlist) (P:kind -> expr -> Prop) : Prop :=
  match el with
  | Enil => True
  | Econs e el => P RV e /\ req_subexprs_list el P
  end.

(* Identify the reduction for an expression where the appropriate subexprs
   are already reduced. *)
Lemma red_for_unstuck : forall k e m,
  not_stuck_for k e m ->
  req_subexprs e (fun k' e' => not_imm_stuck_reduced k' e' m) ->
    match k,e with
    | LV,_ => exists e' m', lred ge en e m e' m'
    | RV,Ecall _ _ _ => exists fd vargs ty, callred ge e fd vargs ty
    | RV,_ => exists e' m', rred e m e' m'
    end.
intros; destruct k; destruct e; simpl in H0; try solve [ destruct H0 ];
inv_imm_stuck H; try inversion CTX; subst; try discriminate; subst;
try solve
 [ exists e' m'; apply RED
 | inversion RED
 | injection E2; intros; subst; inversion H0; subst; inversion H1; subst; try discriminate; subst; inversion RED
 | injection E2; intros; subst; destruct H0 as [R1 R2]; inversion R1; inversion R2; subst; inversion H1; subst; try discriminate; try subst; try inversion RED
 ].

injection E2; intros; subst; destruct H0 as [R1 R2]; inversion R1; subst; inversion H1; subst; try discriminate; subst.
injection E2; intros; subst; destruct H0 as [R1 R2]. clear H R1 H' E2. elimtype False. induction H1.
  simpl in R2. destruct R2 as [R3 R4]; inversion R3; subst. inversion H; subst; try discriminate. subst. inversion RED.
  apply IHcontextlist. apply R2. auto using context.

injection E2; intros; subst; destruct H0 as [R1 R2]; inversion R1; subst; inversion H1; subst; try discriminate; subst. inversion RED.
injection E2; intros; subst; destruct H0 as [R1 R2]. clear H R1 H' E2. elimtype False. assert (RV = RV) by reflexivity. revert H1 H. generalize RV at 1 2. induction 1; intro; subst.
  simpl in R2. destruct R2 as [R3 R4]; inversion R3; subst; inversion H; subst; try discriminate. subst. inversion RED.
  apply IHcontextlist; auto using context. apply R2.

exists fd args ty99. apply RED.

injection E2; intros; subst; destruct H0 as [R1 R2]; inversion R1; subst; inversion H1; subst; try discriminate; subst; inversion RED.
injection E2; intros; subst; destruct H0 as [R1 R2]. clear H R1 H' E2. elimtype False. assert (RV = RV) by reflexivity. revert H1 H. generalize RV at 1 2. induction 1; intro; subst.
  simpl in R2. destruct R2 as [R3 R4]; inversion R3; subst; inversion H; subst; try discriminate. subst. inversion RED.
  apply IHcontextlist; auto using context. apply R2.
Qed.

(* The quantification means that we can't do an induction on the proof of
   not_stuck_for, so we use the expression instead. *)
Scheme expr_ind2 := Induction for expr Sort Prop
with exprlist_ind2 := Induction for exprlist Sort Prop.
Combined Scheme expr_exprlist_ind from expr_ind2, exprlist_ind2.

Lemma check_not_stuck_complete : forall k e m,
  not_stuck_for k e m -> match check_not_stuck k e m with Reduced | Reducible => True | Stuck _ _ _ => False end.

(* Avoid unfolding these - we'll use the completeness lemmas instead. *)
Opaque exec_lred exec_rred exec_callred.

(* Start the induction, get rid of the simple cases. *)
intros. revert e k H.
apply (expr_ind2
 (fun e => forall k, not_stuck_for k e m -> match check_not_stuck k e m with Reduced | Reducible => True | Stuck _ _ _ => False end)
 (fun el => list_not_stuck el m ->
        match check_list_not_stuck el m with
        | Reduced | Reducible => True
        | Stuck _ _ _ => False end));
intros; try destruct k; simpl; auto;
try solve [ match goal with
H : not_stuck_for _ _ _ |- False =>
  inv_imm_stuck H; subst; inversion CTX; subst; try discriminate; subst; inversion RED
end ].

destruct (red_for_unstuck _ _ _ H) as [e' [m' RED]]; simpl; auto.
rewrite (exec_lred_complete _ _ RED). trivial.

assert (SUB:=check_not_stuck_sound LV l m).
specialize H with LV.
destruct (check_not_stuck LV l m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_lred_complete _ _ RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Efield (C x) f ty) e'); auto.

assert (SUB:=check_not_stuck_sound LV l m).
specialize H with LV.
destruct (check_not_stuck LV l m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Evalof (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r m).
specialize H with RV.
destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_lred_complete _ _ RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Ederef (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound LV l m).
specialize H with LV.
destruct (check_not_stuck LV l m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Eaddrof (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r m).
specialize H with RV.
destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Eunop op (C x) ty) e'); auto.

assert (SUB1:=check_not_stuck_sound RV r1 m).
assert (SUB2:=check_not_stuck_sound RV r2 m).
specialize H with RV.
specialize H0 with RV.
destruct (check_not_stuck RV r1 m); auto. destruct (check_not_stuck RV r2 m); auto.
  destruct (red_for_unstuck _ _ _ H1) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Ebinop op r1 (C x) ty) e'); auto.

  destruct (check_not_stuck RV r2 m); auto.
    apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Ebinop op r1 (C x) ty) e'); auto.
    apply H. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Ebinop op (C x) r2 ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r m).
specialize H with RV.
destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Ecast (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r1 m).
specialize H with RV.
destruct (check_not_stuck RV r1 m); auto.
  destruct (red_for_unstuck _ _ _ H2) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H2 k' (fun x => Econdition (C x) r2 r3 ty) e'); auto.

destruct (red_for_unstuck _ _ _ H) as [e' [m' RED]]; simpl; auto.
rewrite (exec_rred_complete RED). trivial.

assert (SUB1:=check_not_stuck_sound LV l m).
assert (SUB2:=check_not_stuck_sound RV r m).
specialize H with LV.
specialize H0 with RV.
destruct (check_not_stuck LV l m); auto. destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H1) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassign l (C x) ty) e'); auto.

  destruct (check_not_stuck RV r m); auto.
    apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassign l (C x) ty) e'); auto.
    apply H. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassign (C x) r ty) e'); auto.


assert (SUB1:=check_not_stuck_sound LV l m).
assert (SUB2:=check_not_stuck_sound RV r m).
specialize H with LV.
specialize H0 with RV.
destruct (check_not_stuck LV l m); auto. destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H1) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassignop op l (C x) tyres ty) e'); auto.

  destruct (check_not_stuck RV r m); auto.
    apply H0. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassignop op l (C x) tyres ty) e'); auto.
    apply H. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Eassignop op (C x) r tyres ty) e'); auto.

assert (SUB:=check_not_stuck_sound LV l m).
specialize H with LV.
destruct (check_not_stuck LV l m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Epostincr id (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r1 m).
specialize H with RV.
destruct (check_not_stuck RV r1 m); auto.
  destruct (red_for_unstuck _ _ _ H1) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H1 k' (fun x => Ecomma (C x) r2 ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r1 m).
assert (SUB2:=check_list_not_stuck_sound rargs m).
specialize H with RV.
destruct (check_not_stuck RV r1 m); auto. destruct (check_list_not_stuck rargs m); auto.
  destruct (red_for_unstuck _ _ _ H1) as [e' [fd [ty' RED]]]; simpl; intuition auto.
  clear -SUB2. induction rargs; simpl; auto. split. apply (SUB2 RV (fun x => Econs x rargs) r1); auto using contextlist.
  apply IHrargs; auto. intros k C e CTX ECTX. apply (SUB2 k (fun x => Econs r1 (C x)) e); auto using contextlist. rewrite ECTX; reflexivity.
  rewrite (exec_callred_complete ge RED); trivial.

  apply H0. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Ecall r1 (C2 x) ty) e2); auto using context.

  destruct (check_list_not_stuck rargs m); auto.

  apply H0. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Ecall r1 (C2 x) ty) e2); auto using context.

  apply H. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Ecall (C2 x) rargs ty) e2); auto using context.

assert (SUB:=check_not_stuck_sound RV r m).
specialize H with RV.
destruct (check_not_stuck RV r m); auto.
  destruct (red_for_unstuck _ _ _ H0) as [e' [m' RED]]; simpl; auto.
  rewrite (exec_rred_complete RED). trivial.

  apply H. intros k' C e' CTX EQ. subst. apply (H0 k' (fun x => Eparen (C x) ty) e'); auto.

assert (SUB:=check_not_stuck_sound RV r1 m).
assert (SUB2:=check_list_not_stuck_sound rl m).
specialize H with RV.
destruct (check_not_stuck RV r1 m); auto. destruct (check_list_not_stuck rl m); auto.
  apply H0. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Econs r1 (C2 x)) e2); auto using contextlist.

  destruct (check_list_not_stuck rl m); auto.
  apply H0. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Econs r1 (C2 x)) e2); auto using contextlist.

  apply H. intros k2 C2 e2 CTX ECTX. subst. apply (H1 k2 (fun x => Econs (C2 x) rl) e2); auto using contextlist.
Qed.

(* Now we can shed the generalisations and go straight for not_stuck. *)

Definition is_stuck (e:expr) (m:mem) : option errmsg :=
match check_not_stuck RV e m with
| Reduced => None
| Reducible => None
| Stuck msg _ _ => Some msg
end.

Theorem is_stuck_sound : forall e m,
  is_stuck e m = None -> not_stuck ge en e m.
unfold is_stuck. intros. assert (H':=check_not_stuck_sound RV e m). destruct (check_not_stuck RV e m).
  intros k C e' CTX ECTX. inversion CTX; intros; subst; inversion H'. subst; auto using not_imm_stuck_inc1.
  intros k C e' CTX ECTX. subst. apply (proj2 H' k C e'); auto.
  discriminate.
Qed.

Theorem is_stuck_complete : forall e m,
  not_stuck ge en e m -> is_stuck e m = None.
intros. unfold is_stuck.
assert (not_stuck_for RV e m).
  intros k C e' CTX ECTX. apply (H k C e' CTX ECTX).

  assert (H1 := check_not_stuck_complete RV e m H0).
  destruct (check_not_stuck RV e m); auto. contradiction.
Qed.

End NOT_STUCK.


Section EXPR.

Hint Constructors Csem.cast_arguments.
Hint Constructors Csem.callred.

Variable ge:Csem.genv.
Variable en:Csem.env.

Open Local Scope error_monad_scope.

(* The pick function identifies a context in which to perform the reduction. *)

Definition exec_estep (pick:expr -> res (Csem.kind * expr * (expr -> expr))) (check_stuck:bool)
 (* ExprState *) (f: function) (r: expr) (k: Csem.cont) (m: mem) : res Csem.state :=
match (if check_stuck then is_stuck ge en r m else None) with
| None =>
  do (ctx,C) <- pick r;
  let (ck,a) := ctx in
  match ck with
  | Csem.LV =>
      do (a',m') <- exec_lred ge en a m;
      OK (Csem.ExprState f (C a') k en m')
  | Csem.RV =>
      match a with
      | Ecall _ _ _ =>
          do (call,ty) <- exec_callred ge a;
          let (fd,vargs) := call in
          OK (Csem.Callstate fd vargs (Csem.Kcall f en C ty k) m)
      | _ =>
          do (a',m') <- exec_rred a m;
          OK (Csem.ExprState f (C a') k en m')
      end
  end
| Some err => Error (MSG "stuck expression: "::err)
end.

Lemma exec_estep_sound : forall pick f r k m s,
  (forall {ck a C}, pick r = OK (ck,a,C) -> Csem.context ck Csem.RV C /\ r = C a) ->
  exec_estep pick true f r k m = OK s -> Csem.estep ge (Csem.ExprState f r k en m) Events.E0 s.
intros. unfold exec_estep in H0.
destruct (is_stuck ge en r m) as [] _eqn; try discriminate.
destruct (pick r) as [ [[ ck a ] C] | ] _eqn; try discriminate.
destruct ck.
  simpl in H0. monadInv H0.
  destruct (H _ _ _ (eq_refl _)). subst.
  apply Csem.step_lred; auto using exec_lred_sound, is_stuck_sound.

  destruct (H _ _ _ (eq_refl _)); subst.
  simpl in H0. destruct a; monadInv H0;
  auto using Csem.step_rred, exec_rred_sound, is_stuck_sound.
  destruct x; inj_subst EQ0; auto using  Csem.step_call, exec_callred_sound, is_stuck_sound.
Qed.

(* We need some equality functions for the next result. *)
Definition val_eq : forall v v':val, {v = v'} + {v <> v'}.
decide equality; auto using Int.eq_dec, Float.eq_dec, ZArith_dec.Z_eq_dec.
Defined.

Scheme Equality for unary_operation.
Scheme Equality for binary_operation.
Scheme Equality for incr_or_decr.

Fixpoint expr_eq (e e':expr) : {e = e'} + {e <> e'}
 with expr_list_eq (el el':exprlist) : {el = el'} + {el <> el'}.
decide equality;
auto using type_eq, val_eq, AST.ident_eq, unary_operation_eq_dec, binary_operation_eq_dec, Int.eq_dec.
apply incr_or_decr_eq_dec.
apply ZArith_dec.Z_eq_dec.

decide equality.
Defined.

Lemma expr_eq_true : forall e, exists p, expr_eq e e = left p.
intro. apply (dec_eq_true expr_eq).
Qed.

(* A weak form of completeness: there is always a context picking function that
   leads to the right result.  However, it's a pretty useless one.  (In
   general a function of the expression alone isn't enough to represent all
   strategies - that is, some strategies may pick different contexts for the
   same expression at different times.) *)

Lemma exec_estep_complete : forall {f r k m s} cs,
  Csem.estep ge (Csem.ExprState f r k en m) Events.E0 s ->
  exists pick, exec_estep pick cs f r k m = OK s.
intros. inversion H; subst.
  exists (fun r => if expr_eq r (C a) then OK (Csem.LV,a,C) else Error (msg "dumb strategy")).
  unfold exec_estep; destruct cs; simpl; try rewrite (is_stuck_complete ge en (C a) m H7); destruct (expr_eq_true (C a)) as [ E0 E ]; rewrite E; simpl;
  rewrite (exec_lred_complete ge en H6); reflexivity.

  exists (fun r => if expr_eq r (C a) then OK (Csem.RV,a,C) else Error (msg "dumb strategy")).
  unfold exec_estep; destruct cs; simpl; try rewrite (is_stuck_complete ge en (C a) m H7); destruct (expr_eq_true (C a)) as [ E0 E ]; rewrite E; simpl;
  destruct a; try (rewrite (exec_rred_complete H6); reflexivity); inversion H6.

  exists (fun r => if expr_eq r (C a) then OK (Csem.RV,a,C) else Error (msg "dumb strategy")).
  unfold exec_estep; destruct cs; simpl; try rewrite (is_stuck_complete ge en (C a) m H7); destruct (expr_eq_true (C a)) as [ E0 E ]; rewrite E; simpl;
  destruct a; try solve [ inversion H6 ]; rewrite (exec_callred_complete ge H6); reflexivity.
Qed.

End EXPR.

Section STEPS.

Hint Constructors Csem.sstep.

Open Local Scope error_monad_scope.

Fixpoint exec_alloc_vars (e:env) (m:mem) (l:list (AST.ident * type)) : env * mem :=
match l with
| nil => (e,m)
| cons (id,ty) vars =>
    let (m1,b1) := Mem.alloc m 0 (sizeof ty) in
    exec_alloc_vars (PTree.set id (b1,ty) e) m1 vars
end.

Lemma exec_alloc_vars_sound : forall {l e m e' m'},
  exec_alloc_vars e m l = (e', m') -> alloc_variables e m l e' m'.
intro; induction l; intros.
  inj_subst H; constructor.

  simpl in H. destruct a. destruct (Mem.alloc m 0 (sizeof t)) as [ m1 b1 ] _eqn.
  econstructor.
    apply Heqp.

    apply IHl. apply H.
Qed.

Lemma exec_alloc_vars_complete : forall {l e m e' m'},
  alloc_variables e m l e' m' -> exec_alloc_vars e m l = (e', m').
induction 1; auto.
simpl. rewrite H. auto.
Qed.

Fixpoint exec_bind_params (e:env) (m:mem) (x:list (AST.ident * type)) (y:list val) : res mem :=
match x,y with
| nil, nil => OK m
| cons (id,ty) params, cons v1 vl =>
    match PTree.get id e with
    | None => Error (msg "missing parameter in env")
    | Some (b,ty') =>
        match type_eq ty ty' with
        | right _ => Error (msg "parameter type mismatch")
        | left _ =>
            match store_value_of_type ty m b Int.zero v1 with
            | None => Error (msg "parameter store failed")
            | Some m1 => exec_bind_params e m1 params vl
            end
        end
     end
| _,_ => Error (msg "parameter list length mismatch")
end.

Lemma exec_bind_params_sound : forall {x y e m m'},
  exec_bind_params e m x y = OK m' -> bind_parameters e m x y m'.
intro; induction x as [ | [id ty] params ]; intros; destruct y; try discriminate.
  inj_subst H; constructor.

  simpl in H. destruct_opt H. destruct v0. destruct (type_eq ty t); try discriminate. subst.
  destruct_opt H.
  econstructor; eauto.
Qed.

Lemma exec_bind_params_complete : forall {x y e m m'},
  bind_parameters e m x y m' -> exec_bind_params e m x y = OK m'.
induction 1; auto.
simpl. rewrite H, H0. destruct (dec_eq_true type_eq ty). rewrite H2. auto.
Qed.

Variable ge:Csem.genv.

Definition ext_malloc vargs m :=
match vargs with
| Vint n::nil =>
    let (m', b) := Mem.alloc m (-4) (Int.unsigned n) in
    match Mem.store AST.Mint32 m' b (-4) (Vint n) with
    | Some m'' => OK (E0, Vptr b Int.zero, m'')
    | None => Error (msg "Bizarre malloc failure")
    end
| _ => Error (msg "Bad arguments for malloc")
end.

Lemma ext_malloc_sound : forall vargs m tr r m',
  ext_malloc vargs m = OK (tr, r, m') ->
  extcall_malloc_sem ge vargs m tr r m'.
Proof with try discriminate H.
intros.
destruct vargs... destruct v... destruct vargs...
simpl in H. destruct (Mem.alloc m (-4) (Int.unsigned i)) as [m1 b] _eqn.
destruct_opt H.
inj_subst H. eauto using extcall_malloc_sem.
Qed.

Require Import BinInt.

Definition ext_free vargs m :=
match vargs with
| Vptr b lo :: nil =>
    match Mem.load AST.Mint32 m b (Int.unsigned lo - 4) with
    | Some (Vint sz) =>
        match ZArith_dec.Z_gt_dec (Int.unsigned sz) 0 with
        | left _ =>
            match Mem.free m b (Int.unsigned lo - 4) (Int.unsigned lo + Int.unsigned sz) with
            | Some m' => OK (E0, Vundef, m')
            | None => Error (msg "free failed")
            end
        | right _ => Error (msg "Block passed to free claims negative size")
        end
    | None => Error (msg "Unable to load size of block for free")
    | _ => Error (msg "Block does not contain suitable size value for free")
    end
| _ => Error (msg "Bad arguments for free")
end.

Lemma ext_free_sound : forall vargs m tr r m',
  ext_free vargs m = OK (tr, r, m') ->
  extcall_free_sem ge vargs m tr r m'.
Proof with try discriminate H.
intros. destruct vargs... destruct v... destruct vargs...
simpl in H.
destruct_opt H. destruct v...
destruct (ZArith_dec.Z_gt_dec (Int.unsigned i0) 0)...
destruct_opt H.
inj_subst H.
eauto using extcall_free_sem.
Qed. 

Variable en:Csem.env.
Variable pick:expr -> res (Csem.kind * expr * (expr -> expr)).

Definition exec_step (s:state) (check_stuck:bool) : res (trace * state) :=
match s with
| ExprState f r k e m =>
    match r with
    | Eval v ty =>
        match k with
        | Kdo k => OK (E0, State f Sskip k e m)
        | Kifthenelse s1 s2 k =>
            do b <- exec_bool_val v ty;
            OK (E0, State f (if b then s1 else s2) k e m)
        | Kwhile1 x s k =>
            do b <- exec_bool_val v ty;
            if b then OK (E0, State f s (Kwhile2 x s k) e m) else OK (E0, State f Sskip k e m)
        | Kdowhile2 x s k =>
            do b <- exec_bool_val v ty;
            if b then OK (E0, State f (Sdowhile x s) k e m) else OK (E0, State f Sskip k e m)
        | Kfor2 a2 a3 s k =>
            do b <- exec_bool_val v ty;
            if b then OK (E0, State f s (Kfor3 a2 a3 s k) e m) else OK (E0, State f Sskip k e m)
        | Kreturn k =>
            do v2 <- exec_cast v ty f.(fn_return);
            match Mem.free_list m (blocks_of_env e) with
            | Some m' => OK (E0, Returnstate v2 (call_cont k) m')
            | None => Error (msg "Failure to free stack frame")
            end
        | Kswitch1 sl k =>
            match v with
            | Vint n => OK (E0, State f (seq_of_labeled_statement (select_switch n sl)) (Kswitch2 k) e m)
            | _ => Error (msg "Switch on non-integer value")
            end
        | _ => Error (msg "Bad continuation from expression")
        end
    | _ => do s' <- exec_estep ge e pick check_stuck f r k m;
           OK (Events.E0, s')
    end
| State f s k e m =>
    match s with
    | Sdo x => OK (E0, ExprState f x (Kdo k) e m)
    | Ssequence s1 s2 => OK (E0, State f s1 (Kseq s2 k) e m)
    | Sskip =>
        match k with
        | Kseq s k => OK (E0, State f s k e m)
        | Kwhile2 x s k => OK (E0, State f (Swhile x s) k e m)
        | Kdowhile1 x s k => OK (E0, ExprState f x (Kdowhile2 x s k) e m)
        | Kfor3 a2 a3 s k => OK (E0, State f a3 (Kfor4 a2 a3 s k) e m)
        | Kfor4 a2 a3 s k => OK (E0, State f (Sfor Sskip a2 a3 s) k e m)
        | Kstop
        | Kcall _ _ _ _ _ =>
            match f.(fn_return) with 
            | Tvoid =>
                match Mem.free_list m (blocks_of_env e) with
                | Some m' => OK (E0, Returnstate Vundef k m')
                | None => Error (msg "Failure to free stack frame")
                end
            | _ => Error (msg "Reached end of function without returning a value")
            end
        | Kswitch2 k => OK (E0, State f Sskip k e m)
        | _ => Error (msg "Bad continuation for Sskip")
        end
    | Scontinue =>
        match k with
        | Kseq s k => OK (E0, State f Scontinue k e m)
        | Kwhile2 x s k => OK (E0, State f (Swhile x s) k e m)
        | Kdowhile1 x s k => OK (E0, ExprState f x (Kdowhile2 x s k) e m)
        | Kfor3 a2 a3 s k => OK (E0, State f a3 (Kfor4 a2 a3 s k) e m)
        | Kswitch2 k => OK (E0, State f Scontinue k e m)
        | _ => Error (msg "Bad continuation for Scontinue")
        end
    | Sbreak =>
        match k with
        | Kseq s k => OK (E0, State f Sbreak k e m)
        | Kwhile2 x s k => OK (E0, State f Sskip k e m)
        | Kdowhile1 x s k => OK (E0, State f Sskip k e m)
        | Kfor3 a2 a3 s k => OK (E0, State f Sskip k e m)
        | Kswitch2 k => OK (E0, State f Sskip k e m)
        | _ => Error (msg "Bad continuation for Sbreak")
        end
   | Sifthenelse a s1 s2 => OK (E0, ExprState f a (Kifthenelse s1 s2 k) e m)
   | Swhile x s => OK (E0, ExprState f x (Kwhile1 x s k) e m)
   | Sdowhile a s => OK (E0, State f s (Kdowhile1 a s k) e m)
   | Sfor Sskip a2 a3 s => OK (E0, ExprState f a2 (Kfor2 a2 a3 s k) e m)
   | Sfor a1 a2 a3 s => OK (E0, State f a1 (Kseq (Sfor Sskip a2 a3 s) k) e m)
   | Sreturn None =>
       match Mem.free_list m (blocks_of_env e) with
       | Some m' => OK (E0, Returnstate Vundef (call_cont k) m')
       | None => Error (msg "Failure to free stack frame")
       end
   | Sreturn (Some x) => OK (E0, ExprState f x (Kreturn k) e m)
   | Sswitch x sl => OK (E0, ExprState f x (Kswitch1 sl k) e m)
   | Slabel _ s => OK (E0, State f s k e m)
   | Sgoto lbl =>
       match find_label lbl f.(fn_body) (call_cont k) with
       | Some (s', k') => OK (E0, State f s' k' e m)
       | None => Error (msg "Missing label for Sgoto")
       end
    end
| Callstate (Internal f) vargs k m =>
    match Coqlib.list_norepet_dec AST.ident_eq (var_names (fn_params f) ++ var_names (fn_vars f)) with
    | left _ =>
        let (e,m1) := exec_alloc_vars empty_env m (f.(fn_params) ++ f.(fn_vars)) in
        do m2 <- exec_bind_params e m1 f.(fn_params) vargs;
        OK (E0, State f f.(fn_body) k e m2)
    | right _ => Error (msg "variable name clash")
    end
| Callstate (External ef targs tres) vargs k m =>
    match classify_external_function ef with
    | EF_malloc =>
        do (trr, m') <- ext_malloc vargs m;
        let (tr, r) := trr in
        OK (tr, Returnstate r k m')
    | EF_free =>
        do (trr, m') <- ext_free vargs m;
        let (tr, r) := trr in
        OK (tr, Returnstate r k m')
(*    | EF_annotation =>*)
    | _ => Error (MSG "Unimplemented 'external' function: " :: CTX ef.(AST.ef_id) :: nil)
    end
| Returnstate v k m =>
    match k with
    | Kcall f e C ty k => OK (E0, ExprState f (C (Eval v ty)) k e m)
    | _ => Error (msg "Bad continuation for Returnstate")
    end
end.



Lemma exec_step_sound : forall s tr s',
  (forall {r ck a C}, pick r = OK (ck,a,C) -> Csem.context ck Csem.RV C /\ r = C a) ->
  exec_step s true = OK (tr, s') -> step ge s tr s'.
unfold exec_step; intros s tr s' PICK_CTX H; destruct s; simpl in H.
  right. destruct s; simpl in H; inj_subst H; auto.
    destruct k; simpl in H; try discriminate; inj_subst H; auto.
      destruct (fn_return f) as [] _eqn; try discriminate; destruct_opt H; inj_subst H.
      apply step_skip_call;simpl;auto.

      destruct (fn_return f) as [] _eqn; try discriminate; destruct_opt H; inj_subst H.
      apply step_skip_call;simpl;auto.

      destruct s1; inj_subst H; auto; apply step_for_start; intro E; discriminate.

      destruct k; try discriminate; inj_subst H; auto.

      destruct k; try discriminate; inj_subst H; auto.

      destruct o.
        inj_subst H; auto.

        destruct_opt H; inj_subst H; auto.

      destruct_opt H. destruct v; inj_subst H; auto.

  destruct r; try (monadInv H; left; eauto using exec_estep_sound); try discriminate.
    right. destruct k; try discriminate; try solve [inj_subst H; auto].
      monadInv H. destruct x.
        apply step_ifthenelse_true; apply (exec_bool_val_sound_true EQ).
        apply step_ifthenelse_false; apply (exec_bool_val_sound_false EQ).

      monadInv H; destruct x; inj_subst EQ0.
        apply step_while_true. apply (exec_bool_val_sound_true EQ).
        apply step_while_false. apply (exec_bool_val_sound_false EQ).

      monadInv H; destruct x; inj_subst EQ0.
        apply step_dowhile_true. apply (exec_bool_val_sound_true EQ).
        apply step_dowhile_false. apply (exec_bool_val_sound_false EQ).

      monadInv H; destruct x; inj_subst EQ0.
        apply step_for_true. apply (exec_bool_val_sound_true EQ).
        apply step_for_false. apply (exec_bool_val_sound_false EQ).

      destruct v; try discriminate. inj_subst H. auto.

      monadInv H. destruct_opt EQ0. inj_subst EQ0. auto using exec_cast_sound.

  right; destruct fd; try discriminate.
    destruct (Coqlib.list_norepet_dec AST.ident_eq (var_names (fn_params f) ++ var_names (fn_vars f))); try discriminate.
    destruct (exec_alloc_vars empty_env m (fn_params f ++ fn_vars f)) as [ e m1 ] _eqn.
    monadInv H; econstructor; eauto. apply (exec_alloc_vars_sound Heqp). apply (exec_bind_params_sound EQ).

  assert (E := eq_refl (external_call e)). unfold external_call in E at 2.
  destruct (classify_external_function e); try discriminate H;
  monadInv H; destruct x; try discriminate; inj_subst EQ0.
    apply ext_malloc_sound in EQ.
    eapply step_external_function.
    rewrite E. assumption.

    apply ext_free_sound in EQ.
    eapply step_external_function.
    rewrite E. assumption.

  right; destruct k; try discriminate.
  inj_subst H; eauto.
Qed.

Definition only_supported_external (s:state) : Prop :=
  match s with
  | Callstate (External ef targs tres) vargs k m =>
      match classify_external_function ef with
      | EF_malloc
      | EF_free => True
      | _ => False
      end
  | _ => True
  end.

Lemma sstep_complete : forall s tr s' check_stuck,
  sstep ge s tr s' ->
  only_supported_external s ->
  exec_step s check_stuck = OK (tr, s').
induction 1; try solve
[ reflexivity
| simpl; rewrite exec_bool_val_true; auto
| simpl; rewrite exec_bool_val_false; auto
].

simpl. destruct H as [E|E]; rewrite E; reflexivity.
simpl. destruct H as [E|E]; rewrite E; reflexivity.
simpl. destruct a1; try reflexivity. congruence.
destruct H as [E|E]; rewrite E; reflexivity.
simpl; rewrite H; reflexivity.
simpl. rewrite (exec_cast_complete H). rewrite H0. reflexivity.
simpl. destruct k; try destruct H; rewrite H0, H1; reflexivity.
destruct H as [E|E]; rewrite E; reflexivity.
simpl; rewrite H; reflexivity.
(* internal functions *)
simpl. destruct (Coqlib.list_norepet_dec AST.ident_eq (var_names (fn_params f) ++ var_names (fn_vars f))).
simpl. rewrite (exec_alloc_vars_complete H0). rewrite (exec_bind_params_complete H1).
reflexivity. contradiction.

(* external functions *)
simpl. unfold external_call in H. destruct (classify_external_function ef); intro X; destruct X.
  (* malloc *)
  inversion H; subst. simpl. rewrite H0, H1. reflexivity.
  (* free *)
  inversion H; subst. simpl. rewrite H0. destruct (ZArith_dec.Z_gt_dec (Int.unsigned sz) 0).
  rewrite H2; reflexivity.
  contradiction.
Qed.

End STEPS.

Theorem exec_step_mostly_complete : forall ge s tr s',
  step ge s tr s' ->
  only_supported_external s ->
  exists pick, exec_step ge pick s true = OK (tr, s').
(* A little too inversion heavy - ideally we'd notice that any estep has a suitable form for
   skipping the sstep cases in exec_step, but this will do. *)
induction 1.
  intro EXT. inversion H; subst.
    destruct (exec_estep_complete ge _ true H) as [strategy ES];
    exists strategy; simpl; rewrite ES; inversion H2; subst; auto.

    destruct (exec_estep_complete ge _ true H) as [strategy ES].
    exists strategy; simpl; rewrite ES. inversion H2; subst; inversion H0; subst; auto.

    destruct (exec_estep_complete ge _ true H) as [strategy ES].
    exists strategy; simpl; rewrite ES. inversion H2; subst; inversion H0; subst; auto.

intro EXT. exists (fun r : expr => OK (RV,r,(fun x : expr => x))).
auto using sstep_complete.
Qed.

(* Initial and final states. *)

Definition make_initial_state (p:program) : res (genv * state) :=
  let ge := Genv.globalenv p in
  match Genv.init_mem p with
  | None => Error (msg "init_mem failed")
  | Some m0 =>
      match Genv.find_symbol ge p.(AST.prog_main) with
      | None => Error (msg "main function symbol not found")
      | Some b =>
          match Genv.find_funct_ptr ge b with
          | None => Error (msg "main function not found")
          | Some f =>
              match type_eq (type_of_fundef f) (Tfunction Tnil (Tint I32 Signed)) with
              | right _ => Error (msg "main function has wrong type")
              | left _ => OK (ge, Callstate f nil Kstop m0)
              end
          end
      end
  end.

Lemma make_initial_state_sound : forall p ge s,
  make_initial_state p = OK (ge,s) -> initial_state p s.
unfold make_initial_state; intros. repeat destruct_opt H.
destruct (type_eq (type_of_fundef v1) (Tfunction Tnil (Tint I32 Signed))) as [] _eqn; try discriminate.
inj_subst H. eauto using initial_state_intro.
Qed.

Lemma make_initial_state_complete : forall p s,
  initial_state p s -> make_initial_state p = OK (Genv.globalenv p, s).
intros; inversion H; subst. unfold ge in *. unfold make_initial_state.
rewrite H0, H1, H2, H3. destruct (dec_eq_true type_eq (Tfunction Tnil (Tint I32 Signed))).
rewrite H4; reflexivity.
Qed.

(* A version that skips the type check to allow programs which appear to take
   arguments, but don't use them. *)
Fixpoint dummy_args (t:typelist) : list val :=
match t with
| Tnil => nil
| Tcons _ tl => Vundef::(dummy_args tl)
end.

Definition make_initial_state_relaxed (p:program) : res (genv * state) :=
  let ge := Genv.globalenv p in
  match Genv.init_mem p with
  | None => Error (msg "init_mem failed")
  | Some m0 =>
      match Genv.find_symbol ge p.(AST.prog_main) with
      | None => Error (msg "main function symbol not found")
      | Some b =>
          match Genv.find_funct_ptr ge b with
          | None => Error (msg "main function not found")
          | Some f =>
              match type_of_fundef f with
              | Tfunction args ret => OK (ge, Callstate f (dummy_args args) Kstop m0)
              | _ => Error (msg "main function has wrong type")
              end
          end
      end
  end.

Definition is_final_state (s:state) : option int :=
match s with
| Returnstate (Vint r) Kstop m => Some r
| _ => None
end.

Lemma is_final_state_sound : forall s r,
  is_final_state s = Some r -> final_state s r.
Proof with try discriminate.
intros; destruct s... destruct res... destruct k... inj_subst H. constructor.
Qed.

Lemma is_final_state_complete : forall s r,
  final_state s r -> is_final_state s = Some r.
intros; inversion H; subst; reflexivity.
Qed.

(* For now we leave whole program execution to Ocaml code. *)


(* The leftmost innermost context picker. *)

(* I had the fixpoints as local definitions, but that looked rather hairy to do
   the proofs with. *)
Fixpoint pick_lmim_expr (e:expr) (to:kind) {struct e} : Csem.kind * expr * (expr -> expr) :=
  match e, to with
  (* First, the expressions whose appropriate subexpressions have been reduced
     enough to move on to the whole expression.  Ecall is handled separately
     below because we need to check the argument list. *)
  | Ederef (Eval _ _) _, _
  | Efield (Eloc _ _ _) _ _, _
  | Evalof (Eloc _ _ _) _, _
  | Eaddrof (Eloc _ _ _) _, _
  | Eunop _ (Eval _ _) _, _
  | Ebinop _ (Eval _ _) (Eval _ _) _, _
  | Ecast (Eval _ _) _, _
  | Econdition (Eval _ _) _ _ _, _
  | Eassign (Eloc _ _ _) (Eval _ _) _, _
  | Eassignop _ (Eloc _ _ _) (Eval _ _) _ _, _
  | Epostincr _ (Eloc _ _ _) _, _
  | Ecomma (Eval _ _) _ _, _
  | Eparen (Eval _ _) _, _
      => (to, e, fun x => x)

  (* Now expressions where a subexpression can be reduced (plus the other case
     for Ecall). *)
  | Ederef e ty, LV => 
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Ederef (C x) ty) end
  | Efield e f ty, LV =>
      match pick_lmim_expr e LV with (k,e,C) => (k, e, fun x => Efield (C x) f ty) end
  | Evalof e ty, RV =>
      match pick_lmim_expr e LV with (k,e,C) => (k, e, fun x => Evalof (C x) ty) end
  | Eaddrof e ty, RV =>
      match pick_lmim_expr e LV with (k,e,C) => (k, e, fun x => Eaddrof (C x) ty) end
  | Eunop op e ty, RV =>
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Eunop op (C x) ty) end
  | Ebinop op (Eval v ty1) e2 ty, RV =>
      match pick_lmim_expr e2 RV with (k,e,C) => (k, e, fun x => Ebinop op (Eval v ty1) (C x) ty) end
  | Ebinop op e1 e2 ty, RV =>
      match pick_lmim_expr e1 RV with (k,e,C) => (k, e, fun x => Ebinop op (C x) e2 ty) end
  | Ecast e ty, RV =>
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Ecast (C x) ty) end
  | Econdition e r2 r3 ty, RV =>
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Econdition (C x) r2 r3 ty) end
  | Eassign (Eloc b ofs ty1) e2 ty, RV =>
      match pick_lmim_expr e2 RV with (k,e,C) => (k, e, fun x => Eassign (Eloc b ofs ty1) (C x) ty) end
  | Eassign e1 e2 ty, RV =>
      match pick_lmim_expr e1 LV with (k,e,C) => (k, e, fun x => Eassign (C x) e2 ty) end
  | Eassignop op (Eloc b ofs ty1) e2 tyres ty, RV =>
      match pick_lmim_expr e2 RV with (k,e,C) => (k, e, fun x => Eassignop op (Eloc b ofs ty1) (C x) tyres ty) end
  | Eassignop op e1 e2 tyres ty, RV =>
      match pick_lmim_expr e1 LV with (k,e,C) => (k, e, fun x => Eassignop op (C x) e2 tyres ty) end
  | Epostincr id e ty, RV =>
      match pick_lmim_expr e LV with (k,e,C) => (k, e, fun x => Epostincr id (C x) ty) end
  | Ecall (Eval v ty0) el ty, _ =>
      match to, pick_lmim_exprlist el with
      | LV, _ (* ??? *)
      | _, None => (to, Ecall (Eval v ty0) el ty, fun x => x)
      | _, Some (k,e,C) => (k, e, fun x => Ecall (Eval v ty0) (C x) ty)
      end
  | Ecall e el ty, RV =>
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Ecall (C x) el ty) end
  | Ecomma e1 e2 ty, RV =>
      match pick_lmim_expr e1 RV with (k,e,C) => (k, e, fun x => Ecomma (C x) e2 ty) end
  | Eparen e ty, RV =>
      match pick_lmim_expr e RV with (k,e,C) => (k, e, fun x => Eparen (C x) ty) end

  (* Finally a catch all. *)
  | _, _ => (to, e, fun x => x)
  end
with pick_lmim_exprlist (e:exprlist) : option (kind * expr * (expr -> exprlist)) :=
  match e with
  | Econs (Eval v ty) el =>
      match pick_lmim_exprlist el with
      | None => None
      | Some (k, e, C) => Some (k, e, fun x => Econs (Eval v ty) (C x))
      end
  | Econs e el =>
      match pick_lmim_expr e RV with (k,e,C) => Some (k,e, fun x => Econs (C x) el) end
  | _ => None
  end.

Definition pick_leftmost_innermost (e:expr) : res (Csem.kind * expr * (expr -> expr)) :=
OK (pick_lmim_expr e RV).


Lemma pick_lmim_ctx :
 (forall e ck ck' e' C, pick_lmim_expr e ck = (ck', e', C) -> e = C e' /\ Cstrategy.leftcontext ck' ck C) /\
 (forall el ck' e' C, pick_lmim_exprlist el = Some (ck', e', C) -> el = C e' /\ Cstrategy.leftcontextlist ck' C).

(* This is rather expensive. :(
   I'd rather use Functional Scheme to get a principle for the fixpoint, but it
   doesn't like the definition. *)

apply expr_exprlist_ind; intros; try destruct ck;
  (* identify the equality hypothesis *)
  match goal with EQH : _ = _ |- _ => rename EQH into EQ end;
  simpl in EQ;

  (* do further pattern matching as necessary *)
  repeat match type of EQ with
    (* oh, this is so tedious *)
    match ?e as expr with Eval _ _ => _ | Evar _ _ => _ | Efield _ _ _ => _ | Evalof _ _ => _ | Ederef _ _ => _ | Eaddrof _ _ => _ | Eunop _ _ _ => _ | Ebinop _ _ _ _ => _ | Ecast _ _ => _ | Econdition _ _ _ _ => _ | Esizeof _ _ => _ | Eassign _ _ _ => _ | Eassignop _ _ _ _ _ => _ | Epostincr _ _ _ => _ | Ecomma _ _ _ => _ | Ecall _ _ _ => _ | Eloc _ _ _ => _ | Eparen _ _ => _ end = _
  => destruct e end;

  (* grab result of (most) recursive calls *)
  try (match type of EQ with context [ pick_lmim_expr ?ex ?ckx ] =>
    destruct (pick_lmim_expr ex ckx) as [[k0 e0] C0] _eqn: EQ0
  end);

  simplify_eq EQ; intros; subst;
  try (destruct (H _ _ _ _ EQ0) as [EQ1 LC1]); try (destruct (H0 _ _ _ _ EQ0) as [EQ2 LC2]);
  try rewrite EQ1; try rewrite EQ2;
  split; try constructor; simpl; auto using Cstrategy.leftcontext, Cstrategy.leftcontextlist.

  destruct (pick_lmim_exprlist rargs) as [ [[k0 e0] C0] | ] _eqn: EQ0;
  simplify_eq EQ; intros; subst; try rewrite (proj1 (H0 _ _ _ (eq_refl _))); auto.

  destruct (pick_lmim_exprlist rargs) as [ [[k0 e0] C0] | ] _eqn: EQ0;
  simplify_eq EQ; intros; subst.
    constructor; simpl; auto.
    apply (proj2 (H0 _ _ _ (eq_refl _))); auto.
    constructor.


  destruct (pick_lmim_exprlist rl) as [ [[k0 e0] C0] | ] _eqn: EQ0;
  simplify_eq EQ; intros; subst; try rewrite (proj1 (H0 _ _ _ (eq_refl _))); auto.

  destruct (pick_lmim_exprlist rl) as [ [[k0 e0] C0] | ] _eqn: EQ0;
  simplify_eq EQ; intros; subst.
    constructor; simpl; auto.
    apply (proj2 (H0 _ _ _ (eq_refl _))); auto.
Qed.

(* This result is misleading - it picks a leftcontext, but not necessarily the
   one that Cstrategy will use, because it evaluates simple subexpressions
   lazily. *)

Theorem pick_leftmost_innermost_leftctx : forall e ck e' C,
  pick_leftmost_innermost e = OK (ck, e', C) -> e = C e' /\ Cstrategy.leftcontext ck RV C.
intros; apply ((proj1 pick_lmim_ctx) _ RV). unfold pick_leftmost_innermost in H. injection H. auto.
Qed.


(* The Cstrategy context picker. *)

Open Local Scope bool_scope.

Fixpoint is_simple (e:expr) : bool :=
  match e with
  | Eloc _ _ _ => true
  | Evar _ _ => true
  | Ederef r _ => is_simple r
  | Efield l1 _ _ => is_simple l1
  | Eval _ _ => true
  | Evalof l _ => is_simple l
  | Eaddrof l _ => is_simple l
  | Eunop _ r1 _ => is_simple r1
  | Ebinop _ r1 r2 _ => is_simple r1 && is_simple r2
  | Ecast r1 _ => is_simple r1
  | Econdition _ _ _ _ => false
  | Esizeof _ _ => true
  | Eassign _ _ _ => false
  | Eassignop _ _ _ _ _ => false
  | Epostincr _ _ _ => false
  | Ecomma _ _ _ => false
  | Ecall _ _ _ => false
  | Eparen _ _ => false
  end.

Lemma is_simple_prop : forall e, is_simple e = true <-> Cstrategy.simple e.
induction e; simpl; intuition.
destruct (andb_prop _ _ H3). auto.
destruct (andb_prop _ _ H3). auto.
Qed.

Definition place_in_ctx (ctx:expr -> expr) (x:expr * (expr -> expr)) : (expr * (expr -> expr)) :=
  match x with (e,C) => (e,fun x => ctx (C x)) end.

Fixpoint find_lmim_effect_expr (e:expr) {struct e} : option (expr * (expr -> expr)) :=
  match e with
  (* Expressions where a subexpression can be reduced (plus the other case
     for Ecall). *)
  | Ederef e ty => 
      option_map (place_in_ctx (fun x => Ederef x ty)) (find_lmim_effect_expr e)
  | Efield e f ty =>
      option_map (place_in_ctx (fun x => Efield x f ty)) (find_lmim_effect_expr e)
  | Evalof e ty =>
      option_map (place_in_ctx (fun x => Evalof x ty)) (find_lmim_effect_expr e)
  | Eaddrof e ty =>
      option_map (place_in_ctx (fun x => Eaddrof x ty)) (find_lmim_effect_expr e)
  | Eunop op e ty =>
      option_map (place_in_ctx (fun x => Eunop op x ty)) (find_lmim_effect_expr e)
  | Ebinop op e1 e2 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Ebinop op x e2 ty) x)
      | None => option_map (place_in_ctx (fun x => Ebinop op e1 x ty)) (find_lmim_effect_expr e2)
      end
  | Ecast e ty =>
      option_map (place_in_ctx (fun x => Ecast x ty)) (find_lmim_effect_expr e)
  | Econdition e1 r2 r3 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Econdition x r2 r3 ty) x)
      | None => Some (e, fun x => x)
      end
  | Eassign e1 e2 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Eassign x e2 ty) x)
      | None =>
          match find_lmim_effect_expr e2 with
          | Some x => Some (place_in_ctx (fun x => Eassign e1 x ty) x)
          | None => Some (e, fun x => x)
          end
      end
  | Eassignop op e1 e2 tyres ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Eassignop op x e2 tyres ty) x)
      | None =>
          match find_lmim_effect_expr e2 with
          | Some x => Some (place_in_ctx (fun x => Eassignop op e1 x tyres ty) x)
          | None => Some (e, fun x => x)
          end
      end
  | Epostincr id e1 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Epostincr id x ty) x)
      | None => Some (e, fun x => x)
      end
  | Ecall e1 el ty =>
      match find_lmim_effect_expr e1 with 
      | Some x => Some (place_in_ctx (fun x => Ecall x el ty) x)
      | None =>
          match find_lmim_effect_exprlist el with
          | Some (e,C) => Some (e, (fun x => Ecall e1 (C x) ty))
          | None => Some (e, fun x => x)
          end
      end
  | Ecomma e1 e2 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Ecomma x e2 ty) x)
      | None => Some (e, fun x => x)
      end
  | Eparen e1 ty =>
      match find_lmim_effect_expr e1 with
      | Some x => Some (place_in_ctx (fun x => Eparen x ty) x)
      | None => Some (e, fun x => x)
      end

  (* Finally a catch all. *)
  | _ => None
  end
with find_lmim_effect_exprlist (e:exprlist) : option (expr * (expr -> exprlist)) :=
  match e with
  | Econs e1 el =>
      match find_lmim_effect_expr e1 with
      | Some (e,C) => Some (e, fun x => Econs (C x) el)
      | None =>
          match find_lmim_effect_exprlist el with
          | None => None
          | Some (e, C) => Some (e, fun x => Econs e1 (C x))
          end
      end
  | _ => None
  end.

Definition pick_cstrategy (e:expr) : res (Csem.kind * expr * (expr -> expr)) :=
  match find_lmim_effect_expr e with
  | Some (e,C) => (* all subexpressions will be simple, so now apply normal strategy. *)
      match pick_lmim_expr e (Cstrategy.expr_kind e) with (k',e',C') => OK (k',e',fun x => C (C' x)) end
  | None => OK (pick_lmim_expr e (Cstrategy.expr_kind e))
  end.

Lemma find_effect_simple : (forall e, find_lmim_effect_expr e = None -> Cstrategy.simple e) /\
                (forall es, find_lmim_effect_exprlist es = None -> Cstrategy.simplelist es).
apply expr_exprlist_ind; intros; simpl in *; auto;
repeat match goal with X:find_lmim_effect_expr ?E = None -> ?Y |- _ => destruct (find_lmim_effect_expr E); simpl in *; auto; try discriminate end.
  destruct (find_lmim_effect_exprlist rargs); simpl in *; auto. destruct p as [e C]. discriminate. discriminate.
  destruct p as [e C]; discriminate.
  destruct (find_lmim_effect_exprlist rl) as [[e C] | ]; simpl in *; auto; try discriminate.
Qed.

Lemma simple_find_effect : (forall e, Cstrategy.simple e -> find_lmim_effect_expr e = None) /\
                (forall es, Cstrategy.simplelist es -> find_lmim_effect_exprlist es = None).
apply expr_exprlist_ind; intros; simpl in *; auto; try contradiction;
try (rewrite (H H0); reflexivity).
rewrite (H (proj1 H1)). rewrite (H0 (proj2 H1)). reflexivity.
rewrite (H (proj1 H1)). rewrite (H0 (proj2 H1)). reflexivity.
Qed.

Scheme leftcontext_ind2 := Minimality for Cstrategy.leftcontext Sort Prop
  with leftcontextlist_ind2 := Minimality for Cstrategy.leftcontextlist Sort Prop.
Combined Scheme leftcontext_leftcontextlist_ind from leftcontext_ind2, leftcontextlist_ind2.

Lemma leftcontext_find_effect :
 (forall k k' C, Cstrategy.leftcontext k k' C -> forall e, Cstrategy.expr_kind e = k -> Cstrategy.simple_side_effect e -> find_lmim_effect_expr (C e) = Some (e, C)) /\
 (forall k C, Cstrategy.leftcontextlist k C -> forall e, Cstrategy.expr_kind e = k -> Cstrategy.simple_side_effect e -> find_lmim_effect_exprlist (C e) = Some (e, C)).
apply leftcontext_leftcontextlist_ind; intros;
  try solve [simpl; rewrite H0; auto].

destruct e; simpl in *; try destruct H0; try rewrite ((proj1 simple_find_effect) _ H0); try reflexivity.
  rewrite ((proj1 simple_find_effect) _ H1); reflexivity.
  rewrite ((proj1 simple_find_effect) _ H1); reflexivity.
  rewrite ((proj2 simple_find_effect) _ H1); reflexivity.

simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite H1; auto.
simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite H1; auto.
simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite H1; auto.
simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite H1; auto.
simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite H1; auto.
Qed.

Open Local Scope error_monad_scope.

Fixpoint exec_steps (n:nat) ge (s:state) : res (trace*state) :=
  match n with
  | O => OK (E0, s)
  | S m =>
      do (tr1,s1) <- exec_steps m ge s;
      do (tr2,s2) <- exec_step ge pick_cstrategy s1 false;
      OK (tr1++tr2,s2)
  end.

Lemma find_effect_ctx :
  (forall k k' C,
  Cstrategy.leftcontext k k' C ->
  forall r r' C',
  find_lmim_effect_expr r = Some (r',C') ->
  C = (fun x => x) \/ (* trick to avoid extensionality *)
  find_lmim_effect_expr (C r) = Some (r',fun x => C (C' x)))
/\
  (forall k C,
  Cstrategy.leftcontextlist k C ->
  forall r r' C',
  find_lmim_effect_expr r = Some (r',C') ->
  find_lmim_effect_exprlist (C r) = Some (r',fun x => C (C' x)))
.
apply leftcontext_leftcontextlist_ind; intros;
  try (right; simpl; destruct (H0 _ _ _ H1) as [E | E]; rewrite E; try rewrite H1; reflexivity).

left; reflexivity.

right; simpl; destruct (H1 _ _ _ H2) as [E | E]; rewrite E; try rewrite H2.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.

right; simpl; destruct (H1 _ _ _ H2) as [E | E]; rewrite E; try rewrite H2.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.

right; simpl; destruct (H1 _ _ _ H2) as [E | E]; rewrite E; try rewrite H2.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.
  rewrite ((proj1 simple_find_effect) _ H); reflexivity.

right; simpl; rewrite (H1 _ _ _ H2). rewrite ((proj1 simple_find_effect) _ H); reflexivity.

simpl. destruct (H0 _ _ _ H1) as [E | E]; rewrite E; try rewrite H1; reflexivity.

simpl. rewrite ((proj1 simple_find_effect) _ H). rewrite (H1 _ _ _ H2). reflexivity.
Qed.
(*
Lemma find_effect_in_ctx :
  (forall k k' C,
  Cstrategy.leftcontext k k' C ->
  forall r r' C1,
  find_lmim_effect_expr r = None ->
  find_lmim_effect_expr (C r) = Some (r',C1) ->
    exists C2, r' = C2 r /\ C = (fun x => C1 (C2 x)) /\ Cstrategy.simple_side_effect r')
/\
  (forall k C,
  Cstrategy.leftcontextlist k C ->
  forall r r' C1,
  find_lmim_effect_expr r = None ->
  find_lmim_effect_exprlist (C r) = Some (r',C1) ->
    exists C2, r' = C2 r /\ C = (fun x => C1 (C2 x)) /\ Cstrategy.simple_side_effect r').
apply leftcontext_leftcontextlist_ind; intros;
 try (
  simpl in H2; destruct (find_lmim_effect_expr (C r)) as [[rx Cx] | ] _eqn:E; try discriminate; injection H2; intros; subst rx C1;
  destruct (H0 _ _ _ H1 E) as [C2 [E1 [E2 S1]]]; exists C2; rewrite E2; intuition
 ).

congruence.

simpl in H2. destruct (find_lmim_effect_expr (C r)) as [[rx Cx] | ] _eqn:E; try discriminate. injection H2; intros. subst rx C1.
  destruct (H0 _ _ _ H1 E) as [C2 [E1 [E2 S1]]]. exists C2. rewrite E2. intuition.
This isn't right - can't have a general leftcontext
*)

(* left contexts which yield simple expressions and fully reduce to values on
   the left.  Note the difference from leftcontext for binop. *)

Inductive simple_leftcontext: kind -> kind -> (expr -> expr) -> Prop :=
  | slctx_top: forall k,
      simple_leftcontext k k (fun x => x)
  | slctx_deref: forall k C ty,
      simple_leftcontext k RV C -> simple_leftcontext k LV (fun x => Ederef (C x) ty)
  | slctx_field: forall k C f ty,
      simple_leftcontext k LV C -> simple_leftcontext k LV (fun x => Efield (C x) f ty)
  | slctx_rvalof: forall k C ty,
      simple_leftcontext k LV C -> simple_leftcontext k RV (fun x => Evalof (C x) ty)
  | slctx_addrof: forall k C ty,
      simple_leftcontext k LV C -> simple_leftcontext k RV (fun x => Eaddrof (C x) ty)
  | slctx_unop: forall k C op ty,
      simple_leftcontext k RV C -> simple_leftcontext k RV (fun x => Eunop op (C x) ty)
  | slctx_binop_left: forall k C op e2 ty,
      Cstrategy.simple e2 ->
      simple_leftcontext k RV C -> simple_leftcontext k RV (fun x => Ebinop op (C x) e2 ty)
  | slctx_binop_right: forall k C op v1 ty1 ty,
      simple_leftcontext k RV C ->
      simple_leftcontext k RV (fun x => Ebinop op (Eval v1 ty1) (C x) ty)
  | slctx_cast: forall k C ty,
      simple_leftcontext k RV C -> simple_leftcontext k RV (fun x => Ecast (C x) ty)
.

Definition not_reduced (e:expr) : Prop :=
  match e with
  | Eval _ _ => False
  | Eloc _ _ _ => False
  | _ => True
  end.

Lemma pick_simple : forall {k1 k2 C r k' r' C'},
  simple_leftcontext k1 k2 C ->
  not_reduced r ->
  pick_lmim_expr r k1 = (k', r', C') ->
  C = (fun x => x) \/  (* avoid extensionality concerns *)
  pick_lmim_expr (C r) k2 = (k', r', fun x => C (C' x)).
intros. induction H;
try (
  right; simpl; inversion H; subst; solve
  [ destruct r; try destruct H0; rewrite H1; reflexivity
  | destruct (IHsimple_leftcontext H1) as [E|E];
    [ assert (X : Eval Vundef Tvoid = (fun x => x) (Eval Vundef Tvoid)) by reflexivity; rewrite <- E in X; discriminate
    | rewrite E; reflexivity
    ]
  ]).

left; reflexivity.

  right; simpl; inversion H2; subst; solve
  [ destruct r; try destruct H0; rewrite H1; reflexivity
  | destruct (IHsimple_leftcontext H1) as [E|E];
    [ assert (X : Eval Vundef Tvoid = (fun x => x) (Eval Vundef Tvoid)) by reflexivity; rewrite <- E in X; discriminate
    | rewrite E; reflexivity
    ]
  ].

Qed.

Lemma simple_leftcontext_simple : forall {k k' C r},
  simple_leftcontext k k' C ->
  Cstrategy.simple r ->
  Cstrategy.simple (C r).
induction 1; simpl; intuition.
Qed.

Lemma simple_leftcontext_kind : forall {k k' C r},
  simple_leftcontext k k' C ->
  Cstrategy.expr_kind r = k ->
  Cstrategy.expr_kind (C r) = k'.
inversion 1; subst; auto.
Qed.

Lemma pick_simple' : forall k2 C r k' r' C',
  simple_leftcontext (Cstrategy.expr_kind r) k2 C ->
  Cstrategy.simple r ->
  not_reduced r ->
  pick_cstrategy r = OK (k', r', C') ->
  C = (fun x => x) \/
  pick_cstrategy (C r) = OK (k', r', fun x => C (C' x)).
intros. unfold pick_cstrategy in *.
rewrite ((proj1 simple_find_effect) r H0) in H2. injection H2; intro E.
rewrite ((proj1 simple_find_effect) (C r)).
destruct (pick_simple H H1 E) as [ E2 | E2 ].
 left; apply E2.
 right; rewrite (simple_leftcontext_kind H); auto; rewrite E2; reflexivity.
 apply (simple_leftcontext_simple H H0).
Qed.

Lemma simple_leftcontext_compose : forall k2 k3 C2,
  simple_leftcontext k2 k3 C2 ->
  forall k1 C1,
  simple_leftcontext k1 k2 C1 ->
  simple_leftcontext k1 k3 (fun x => C2 (C1 x)).
intros k2 k3 C2 H2. induction H2; intros; try (constructor; auto).
inversion H; subst; auto.
Qed.

Lemma simple_leftcontext_leftcontext : forall k k' C,
  simple_leftcontext k k' C -> Cstrategy.leftcontext k k' C.
intros; induction H; constructor; auto. simpl; trivial.
Qed.

Inductive simple_estep (ge:genv) : state -> trace -> state -> Prop :=

  | sestep_lred: forall C f a k e m a' m' ck,
      lred ge e a m a' m' ->
      Cstrategy.simple a ->
      simple_leftcontext LV ck C ->
      simple_estep ge (ExprState f (C a) k e m)
         E0 (ExprState f (C a') k e m')

  | sestep_rred: forall C f a k e m a' m' ck,
      rred a m a' m' ->
      Cstrategy.simple a ->
      simple_leftcontext RV ck C ->
      simple_estep ge (ExprState f (C a) k e m)
         E0 (ExprState f (C a') k e m').

Lemma eval_simple_is_simple : forall ge e m,
  (forall r v, Cstrategy.eval_simple_rvalue ge e m r v -> Cstrategy.simple r) /\
  (forall l b o, Cstrategy.eval_simple_lvalue ge e m l b o -> Cstrategy.simple l).
intros. apply Cstrategy.eval_simple_rvalue_lvalue_ind; simpl; auto.
Qed.

Lemma eval_simple_list_is_simplelist : forall ge e m rs tys vs,
  Cstrategy.eval_simple_list ge e m rs tys vs ->
  Cstrategy.simplelist rs.
induction 1; simpl.
  trivial.
  split. eapply (proj1 (eval_simple_is_simple ge e m)). apply H. apply IHeval_simple_list.
Qed.

Lemma simple_steps : forall ge f k e m,
  (forall r v, Cstrategy.eval_simple_rvalue ge e m r v ->
   forall C ck, simple_leftcontext RV ck C ->
   Smallstep.star simple_estep ge (ExprState f (C r) k e m) E0 (ExprState f (C (Eval v (typeof r))) k e m))
/\
  (forall r b o, Cstrategy.eval_simple_lvalue ge e m r b o ->
   forall C ck, simple_leftcontext LV ck C ->
   Smallstep.star simple_estep ge (ExprState f (C r) k e m) E0 (ExprState f (C (Eloc b o (typeof r))) k e m)).
intros. apply Cstrategy.eval_simple_rvalue_lvalue_ind.

eauto using Smallstep.star.

intros. assert (HC:simple_leftcontext LV ck (fun x => C (Evalof x ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_rred with (ck:=ck). rewrite H1 in *. apply red_rvalof. apply H2. simpl; trivial. apply H3. reflexivity.

intros. assert (HC:simple_leftcontext LV ck (fun x => C (Eaddrof x ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_rred with (ck:=ck). apply red_addrof. simpl; trivial. assumption. reflexivity.

intros. assert (HC:simple_leftcontext RV ck (fun x => C (Eunop op x ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_rred with (ck:=ck). apply red_unop; auto. simpl; trivial. assumption. reflexivity.

intros.
eapply Smallstep.star_right with (s2 := (ExprState f (C (Ebinop op (Eval v1 (typeof r1)) (Eval v2 (typeof r2)) ty)) k e m)).
assert (HC1:simple_leftcontext RV ck (fun x => C (Ebinop op x r2 ty))). eapply simple_leftcontext_compose. apply H4. apply slctx_binop_left. eapply (proj1 (eval_simple_is_simple _ _ _)); eassumption. constructor. 
assert (HC2:simple_leftcontext RV ck (fun x => C (Ebinop op (Eval v1 (typeof r1)) x ty))). eapply simple_leftcontext_compose. apply H4. apply slctx_binop_right. constructor.
eapply (Smallstep.star_trans (H0 _ _ HC1) (H2 _ _ HC2)).
reflexivity.
apply sestep_rred with (ck:=ck). apply red_binop; auto. simpl; trivial. tauto. assumption. reflexivity.

intros. assert (HC:simple_leftcontext RV ck (fun x => C (Ecast x ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_rred with (ck:=ck). apply red_cast; auto. simpl; trivial. assumption. reflexivity.

intros. apply Smallstep.star_one. apply sestep_rred with (ck:=ck). apply red_sizeof. simpl; trivial. assumption.

(* lvalues *)

eauto using Smallstep.star.

intros. apply Smallstep.star_one. apply sestep_lred with (ck:=ck). apply red_var_local; auto. simpl;trivial. assumption.

intros. apply Smallstep.star_one. apply sestep_lred with (ck:=ck). apply red_var_global; auto. simpl;trivial. assumption.

intros. assert (HC:simple_leftcontext RV ck (fun x => C (Ederef x ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_lred with (ck:=ck). apply red_deref; auto. simpl; trivial. assumption. reflexivity.

intros. assert (HC:simple_leftcontext LV ck (fun x => C (Efield x f0 ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_lred with (ck:=ck). rewrite H1. apply red_field_struct; auto. simpl; trivial. assumption. reflexivity.


intros. assert (HC:simple_leftcontext LV ck (fun x => C (Efield x f0 ty))) by eauto using simple_leftcontext_compose, simple_leftcontext.
eapply (Smallstep.star_right _ _ (H0 _ _ HC)).
apply sestep_lred with (ck:=ck). rewrite H1. apply red_field_union; auto. simpl; trivial. assumption. reflexivity.
Qed.

Inductive exec_star (ge:genv) : state -> trace -> state -> Prop :=
| es_refl : forall s, exec_star ge s E0 s
| es_step : forall s tr1 s1 tr2 s' tr,
    exec_step ge pick_cstrategy s false = OK (tr1,s1) ->
    exec_star ge s1 tr2 s' ->
    tr = tr1 ** tr2 ->
    exec_star ge s tr s'.

Inductive estep_ctx : kind -> (expr -> expr) -> Prop :=
| esc_simple : forall k, estep_ctx k (fun x:expr => x)
| esc_condition : forall C r2 r3 ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx RV (fun x => C (Econdition x r2 r3 ty))
| esc_assign_left : forall C r ty,
    Cstrategy.leftcontext RV RV C ->
    Cstrategy.simple r ->
    estep_ctx LV (fun x => C (Eassign x r ty))
| esc_assign_right : forall C b ofs lty ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx RV (fun x => C (Eassign (Eloc b ofs lty) x ty))
| esc_assignop_left : forall C op r tyres ty,
    Cstrategy.leftcontext RV RV C ->
    Cstrategy.simple r ->
    estep_ctx LV (fun x => C (Eassignop op x r tyres ty))
| esc_assignop_right : forall C op b ofs lty tyres ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx RV (fun x => C (Eassignop op (Eloc b ofs lty) x tyres ty))
| esc_postincr : forall C id ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx LV (fun x => C (Epostincr id x ty))
| esc_comma : forall C r2 ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx RV (fun x => C (Ecomma x r2 ty))
| esc_paren : forall C ty,
    Cstrategy.leftcontext RV RV C ->
    estep_ctx RV (fun x => C (Eparen x ty))
| esc_call_left : forall C rs ty,
    Cstrategy.leftcontext RV RV C ->
    Cstrategy.simplelist rs ->
    estep_ctx RV (fun x => C (Ecall x rs ty))
| esc_call_arg : forall C v vty rs C' ty,
    Cstrategy.leftcontext RV RV C ->
    Cstrategy.simplelist rs ->
    estep_ctx_arg rs C' ->
    estep_ctx RV (fun x => C (Ecall (Eval v vty) (C' x) ty))
with estep_ctx_arg : exprlist -> (expr -> exprlist) -> Prop :=
| esca_red : forall v ty tl C,
    estep_ctx_arg tl C ->
    estep_ctx_arg (Econs (Eval v ty) tl) (fun x => Econs (Eval v ty) (C x))
| esca_here : forall r tl,
    estep_ctx_arg (Econs r tl) (fun x => Econs x tl).

Lemma find_estep_ctx : forall a C k' a' C',
  Cstrategy.simple a ->
  match a with Eval _ _ | Eloc _ _ _ => False | _ => True end ->
  estep_ctx (Cstrategy.expr_kind a) C ->
  pick_lmim_expr a (Cstrategy.expr_kind a) = (k',a',C') ->
  C = (fun x:expr => x) \/
  pick_cstrategy (C a) = OK (k',a',fun x => C (C' x)).
intros until C'. intros SIMPLE REDUCIBLE ECTX PICK. inversion ECTX.

left. reflexivity.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
simpl; tauto.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
simpl; tauto.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
simpl; tauto.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
simpl; tauto.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
simpl; tauto.

right. unfold pick_cstrategy. rewrite ((proj1 leftcontext_find_effect) _ _ _ H0); auto.
simpl. assert (X:pick_lmim_exprlist (C'0 a) = Some (k', a', fun x => (C'0 (C' x)))).
clear H1 H3.
induction H2.
 simpl. rewrite IHestep_ctx_arg. reflexivity.
 simpl; destruct a; try destruct REDUCIBLE; try discriminate H; simpl Cstrategy.expr_kind in PICK; rewrite PICK; reflexivity.
rewrite X. reflexivity.
simpl. intuition. clear -SIMPLE H1 H2. induction H2; simpl in *; intuition.
Qed.

Lemma not_Eval : forall r (T:Type) (P:val -> type -> T) (Q:T), match r with Eval _ _ => False | _ => True end ->
  match r with Eval v ty => P v ty | _ => Q end = Q.
intros; destruct r; auto. destruct H.
Qed.

Lemma use_not_reduced : forall r (T:Type) (P:val -> type -> T) (Q:Values.block -> int -> type -> T) (R:T), match r with Eval _ _ | Eloc _ _ _ => False | _ => True end ->
  match r with Eval v ty => P v ty | Eloc b ofs ty => Q b ofs ty | _ => R end = R.
intros; destruct r; auto; destruct H.
Qed.

Lemma pick_ctx_lred : forall {ge e k r m r' m' C},
  simple_leftcontext LV k C ->
  lred ge e r m r' m' ->
  pick_lmim_expr (C r) k = (LV,r,C).
intros. induction H; try (
  simpl; rewrite use_not_reduced;
  [ rewrite IHsimple_leftcontext; reflexivity
  | inversion H; subst; try apply I; inversion H0; subst; apply I
]).

inversion H0; subst; reflexivity.

simpl. rewrite use_not_reduced. rewrite IHsimple_leftcontext; reflexivity.
inversion H1; subst; try apply I; inversion H0; subst; apply I.
Qed.

Lemma pick_ctx_lred' : forall {ge e k r m r' m' C},
  simple_leftcontext LV k C ->
  lred ge e r m r' m' ->
  pick_cstrategy (C r) = OK (LV,r,C).
intros. unfold pick_cstrategy. rewrite (proj1 simple_find_effect).
  rewrite (simple_leftcontext_kind H). rewrite (pick_ctx_lred H H0). reflexivity. apply (Cstrategy.lred_kind _ _ _ _ _ _ H0).
  apply (simple_leftcontext_simple H). inversion H0; simpl; auto.
Qed.

Lemma pick_ctx_rred : forall {k r m r' m' C},
  simple_leftcontext RV k C ->
  Cstrategy.simple r ->
  rred r m r' m' ->
  pick_lmim_expr (C r) k = (RV,r,C).
intros. induction H; try (
  simpl; rewrite use_not_reduced;
  [ rewrite IHsimple_leftcontext; reflexivity
  | inversion H; subst; try apply I; inversion H1; subst; apply I
]).

inversion H1; subst; inversion H0; reflexivity.

simpl. rewrite use_not_reduced. rewrite IHsimple_leftcontext; reflexivity.
inversion H2; subst; try apply I; inversion H1; subst; apply I.
Qed.

Lemma pick_ctx_rred' : forall {k r m r' m' C},
  simple_leftcontext RV k C ->
  Cstrategy.simple r ->
  rred r m r' m' ->
  pick_cstrategy (C r) = OK (RV,r,C).
intros. unfold pick_cstrategy. rewrite (proj1 simple_find_effect).
  rewrite (simple_leftcontext_kind H). rewrite (pick_ctx_rred H H0 H1). reflexivity. apply (Cstrategy.rred_kind _ _ _ _ H1).
  apply (simple_leftcontext_simple H). apply H0.
Qed.

Lemma simple_exec : forall C ge f r tr r' k e m,
  estep_ctx (Cstrategy.expr_kind r) C ->
  simple_estep ge (ExprState f r k e m) tr (ExprState f r' k e m) ->
  exec_step ge pick_cstrategy (ExprState f (C r) k e m) false = OK (tr, (ExprState f (C r') k e m)).
intros. inversion H0; subst.
simpl. rewrite (not_Eval (C (C0 a))).
unfold exec_estep. destruct find_estep_ctx with (a:=C0 a) (C:=C) (k':=Cstrategy.expr_kind a) (a':=a) (C':=C0).
  apply (simple_leftcontext_simple H11 H10).
  inversion H11; simpl; trivial. inversion H7; simpl; trivial.
  assumption.
  rewrite (Cstrategy.lred_kind _ _ _ _ _ _ H7); auto. rewrite (simple_leftcontext_kind H11); auto. apply (pick_ctx_lred H11 H7).
  apply (Cstrategy.lred_kind _ _ _ _ _ _ H7).

rewrite H1. rewrite (pick_ctx_lred' H11 H7). simpl. rewrite (exec_lred_complete _ _ H7). reflexivity.

rewrite H1. simpl. rewrite (Cstrategy.lred_kind _ _ _ _ _ _ H7); auto. rewrite (exec_lred_complete _ _ H7). reflexivity.

inversion H; auto; try match goal with X:Cstrategy.leftcontext _ _ ?C |- _ => solve [inversion X; auto] end. inversion H11; auto. inversion H7; auto.

(* rred *)

simpl. rewrite not_Eval.
unfold exec_estep. destruct find_estep_ctx with (a:=C0 a) (C:=C) (k':=Cstrategy.expr_kind a) (a':=a) (C':=C0).
  apply (simple_leftcontext_simple H11 H10).
  inversion H11; inversion H7; simpl; trivial.
  assumption.
  rewrite (Cstrategy.rred_kind _ _ _ _ H7); auto. rewrite (simple_leftcontext_kind H11); auto. apply (pick_ctx_rred H11 H10 H7).
  apply (Cstrategy.rred_kind _ _ _ _ H7).

rewrite H1. rewrite (pick_ctx_rred' H11 H10 H7). simpl. rewrite (exec_rred_complete H7). inversion H7; reflexivity.

rewrite H1. simpl. rewrite (Cstrategy.rred_kind _ _ _ _ H7); auto. rewrite (exec_rred_complete H7). inversion H7; reflexivity.

inversion H; auto; try match goal with X:Cstrategy.leftcontext _ _ ?C |- _ => solve [inversion X; auto] end. inversion H11; auto.
inversion H7; auto.
Qed.

Lemma simple_estep_pure : forall {ge f r k e m tr s'},
  simple_estep ge (ExprState f r k e m) tr s' ->
  exists r', s' = ExprState f r' k e m.
intros. inversion H; subst.
  inversion H7; try destruct H8; econstructor; reflexivity.
  inversion H7; subst; try destruct H8; econstructor; reflexivity.
Qed.

Lemma simple_estep_kind : forall {ge f r r' k e m tr},
  simple_estep ge (ExprState f r k e m) tr (ExprState f r' k e m) ->
  Cstrategy.expr_kind r = Cstrategy.expr_kind r'.
intros. inversion H; subst.
  repeat rewrite (simple_leftcontext_kind H10); auto; induction H6; auto.
  repeat rewrite (simple_leftcontext_kind H10); auto; induction H6; try destruct H9; auto.
Qed.

Lemma exec_star_right : forall ge s tr1 s1,
  exec_star ge s tr1 s1 ->
  forall tr2 tr s',
  exec_step ge pick_cstrategy s1 false = OK (tr2,s') ->
  tr = tr1 ** tr2 ->
  exec_star ge s tr s'.
induction 1.
  intros. rewrite H0. eapply es_step. apply H. apply es_refl. rewrite E0_right. reflexivity.
  intros. apply es_step with (tr1:=tr1) (s1:=s1) (tr2:=tr2**tr0); auto.
  apply IHexec_star with (tr3:=tr0); auto. rewrite H3. rewrite H1. apply Eapp_assoc.
Qed.

Lemma exec_star_trans : forall ge s1 tr1 s2,
  exec_star ge s1 tr1 s2 ->
  forall tr2 s3 tr,
  exec_star ge s2 tr2 s3 ->
  tr = tr1 ** tr2 ->
  exec_star ge s1 tr s3.
induction 1.
  intros. rewrite H0. apply H.
  intros. eapply es_step. apply H. apply IHexec_star with (tr3:=tr0). assumption.
   reflexivity. rewrite H3. rewrite H1. apply Eapp_assoc.
Qed.

Lemma simple_execs : forall C ge f r tr r' k e m,
  estep_ctx (Cstrategy.expr_kind r) C ->
  Smallstep.star simple_estep ge (ExprState f r k e m) tr (ExprState f r' k e m) ->
  exec_star ge (ExprState f (C r) k e m) tr (ExprState f (C r') k e m).
intros. assert (ExprState f r k e m = ExprState f r k e m) by reflexivity. dependent induction H0.
  apply es_refl.

  destruct (simple_estep_pure H2) as [r1 E1]. subst s2.
  apply es_step with (tr1:=t1) (s1:=ExprState f (C r1) k e m) (tr2:=t2).
  apply simple_exec; auto.
  apply IHstar; auto.
  rewrite <- (simple_estep_kind H2); assumption.
  apply H1.
Qed.

Lemma eval_simple_rvalue_rv : forall {ge e m r v},
  Cstrategy.eval_simple_rvalue ge e m r v ->
  Cstrategy.expr_kind r = RV.
inversion 1; reflexivity.
Qed.

Lemma eval_simple_lvalue_lv : forall {ge e m r b ofs},
  Cstrategy.eval_simple_lvalue ge e m r b ofs ->
  Cstrategy.expr_kind r = LV.
inversion 1; reflexivity.
Qed.
(*
Inductive expr_val_list : exprlist -> list val -> exprlist -> Prop :=
| ev_nil : expr_val_list Enil nil Enil
| ev_cons : forall r rt v vt rvt,
    expr_val_list rt vt rvt ->
    expr_val_list (Econs r rt) (v::vt) (Econs (Eval v (typeof r)) rt).
*)
(*
Inductive partially_eval_args : exprlist ... -> list val -> Prop :=
| pea_nil : partially_eval_args Enil nil
| pea_vcons : forall v ty rargs vargs,
    partially_eval_args rargs vargs ->
    partially_eval_args (Econs (Eval v ty) rargs) (v::vargs)
| pea_rcons : forall r v ty rargs vargs,
    partially_eval_args rargs vargs ->
    match r with Eval _ _ => False | _ => True end ->
    partially_eval_args (Econs r rargs) (v::vargs)
*)

Lemma eval_simple_list'_is_simple : forall {ge e m rargs vargs},
  Cstrategy.eval_simple_list' ge e m rargs vargs ->
  Cstrategy.simplelist rargs.
induction 1.
  simpl; trivial.
  split. apply (proj1 (eval_simple_is_simple ge e m) _ _ H).
  apply IHeval_simple_list'.
Qed.

Lemma simple_args_steps : forall ge e m rargs vargs f k v vty ty C,
  Cstrategy.leftcontext RV RV C ->
  Cstrategy.eval_simple_list' ge e m rargs vargs ->
  exec_star ge (ExprState f (C (Ecall (Eval v vty) rargs ty)) k e m) E0 (ExprState f (C (Ecall (Eval v vty) (Cstrategy.rval_list vargs rargs) ty)) k e m).
intros until C. intros Hctx EVAL.
assert (H:Cstrategy.exprlist_all_values Enil) by (simpl;trivial). revert H.
change rargs with (Cstrategy.exprlist_app Enil rargs) at 1.
change (Cstrategy.rval_list vargs rargs) with (Cstrategy.exprlist_app Enil (Cstrategy.rval_list vargs rargs)).
generalize Enil. induction EVAL.
  intros. apply es_refl.

  intros redargs Hredargs.
  apply exec_star_trans with (tr1:=E0) (s2:=ExprState f (C (Ecall (Eval v vty) (Cstrategy.exprlist_app redargs (Econs (Eval v0 (typeof r)) rl)) ty)) k e m) (tr2:=E0).
    apply simple_execs with (C:=(fun x => C (Ecall (Eval v vty) (Cstrategy.exprlist_app redargs (Econs x rl)) ty))).
    rewrite (eval_simple_rvalue_rv H). eapply esc_call_arg with (C:=C). assumption. instantiate (1:=Cstrategy.exprlist_app redargs (Econs r rl)).
    clear -H EVAL Hredargs. induction redargs.
      split. apply (proj1 (eval_simple_is_simple ge e m) _ _ H). apply (eval_simple_list'_is_simple EVAL).
      split. simpl in Hredargs. destruct r1; try destruct Hredargs; simpl; trivial. apply IHredargs; auto.
      destruct r1; simpl in Hredargs; try apply Hredargs; destruct Hredargs.
    clear -Hredargs. induction redargs. simpl. apply esca_here.
      simpl. destruct r1; try destruct Hredargs. apply esca_red. apply IHredargs. apply Hredargs.

    apply (proj1 (simple_steps ge f k e m)) with (C:=fun x => x) (ck:=RV); auto.
    constructor.

    simpl.
    change (Econs (Eval v0 (typeof r)) rl) with (Cstrategy.exprlist_app (Econs (Eval v0 (typeof r)) Enil) rl).
    change (Econs (Eval v0 (typeof r)) (Cstrategy.rval_list vl rl)) with (Cstrategy.exprlist_app (Econs (Eval v0 (typeof r)) Enil) (Cstrategy.rval_list vl rl)).
    rewrite <- Cstrategy.exprlist_app_assoc. rewrite <- Cstrategy.exprlist_app_assoc.
    apply IHEVAL.

    clear -Hredargs. induction redargs; simpl in *. trivial. destruct r1; try contradiction. auto.

    reflexivity.
Qed.

Lemma eval_simple_list_decompose : forall {ge e m rargs targs vargs},
  Cstrategy.eval_simple_list ge e m rargs targs vargs ->
  exists vl,
    Cstrategy.eval_simple_list' ge e m rargs vl /\
    cast_arguments (Cstrategy.rval_list vl rargs) targs vargs.
induction 1.
  exists (@nil val). split; constructor.
  destruct IHeval_simple_list as [vl' [H' Hc']].
  exists (v'::vl'). split; constructor; auto.
Qed.

Lemma pick_no_vargs : forall vl rargs,
  pick_lmim_exprlist (Cstrategy.rval_list vl rargs) = None.
induction vl; auto.
intro. case rargs; auto.
intros; simpl. rewrite IHvl. reflexivity.
Qed.

Lemma cstrategy_estep_complete : forall ge s tr s',
  Cstrategy.estep ge s tr s' ->
  exec_star ge s tr s'.
induction 1.

apply simple_execs with (C:=fun x:expr => x); eauto using estep_ctx.
rewrite H1. apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); auto using simple_leftcontext.

eapply exec_star_right. apply simple_execs with (C:=(fun x => C (Econdition x r2 r3 ty))).
  rewrite (eval_simple_rvalue_rv H0). apply esc_condition; auto.
  apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite (exec_bool_val_true H1). simpl.
  rewrite H2. destruct (dec_eq_true type_eq ty). rewrite H3. simpl. reflexivity.
  reflexivity. simpl; trivial. inversion H; auto. reflexivity.

eapply exec_star_right. apply simple_execs with (C:=(fun x => C (Econdition x r2 r3 ty))).
  rewrite (eval_simple_rvalue_rv H0). apply esc_condition; auto.
  apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite (exec_bool_val_false H1). simpl.
  rewrite H2. destruct (dec_eq_true type_eq ty). rewrite H3. simpl. reflexivity.
  reflexivity. simpl; trivial. inversion H; auto. reflexivity.

(* Eassign *)
eapply exec_star_right. 
  eapply exec_star_trans.
    apply simple_execs with (C:=(fun x => C (Eassign x r ty))).
    rewrite (eval_simple_lvalue_lv H0). apply esc_assign_left; auto. eapply (proj1 (eval_simple_is_simple _ _ _)); apply H1.
    apply (proj2 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=LV); eauto using simple_leftcontext.

    apply simple_execs with (C:=(fun x => C (Eassign (Eloc b ofs (typeof l)) x ty))).
    rewrite (eval_simple_rvalue_rv H1). apply esc_assign_right; auto.
    apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
    reflexivity.

  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite H4. destruct assert_type_true with (ty:=typeof l). rewrite H5. simpl.
  rewrite (exec_cast_complete H2). simpl.
  rewrite H3. simpl.
  reflexivity.
  reflexivity. simpl; tauto. inversion H; auto. reflexivity.

(* Eassignop *)
eapply exec_star_right. 
  eapply exec_star_trans.
    apply simple_execs with (C:=(fun x => C (Eassignop op x r tyres ty))).
    rewrite (eval_simple_lvalue_lv H0). apply esc_assignop_left; auto. eapply (proj1 (eval_simple_is_simple _ _ _)); apply H2.
    apply (proj2 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=LV); eauto using simple_leftcontext.

    apply simple_execs with (C:=(fun x => C (Eassignop op (Eloc b ofs (typeof l)) x tyres ty))).
    rewrite (eval_simple_rvalue_rv H2). apply esc_assignop_right; auto.
    apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
    reflexivity.

  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite H1.
  rewrite H6. destruct assert_type_true with (ty:=typeof l). rewrite H7. simpl.
  rewrite H3; simpl.
  rewrite (exec_cast_complete H4); simpl.
  rewrite H5; simpl.
  reflexivity.
  reflexivity. simpl; tauto. inversion H; auto. reflexivity.

(* Epostincr *)
eapply exec_star_right. apply simple_execs with (C:=(fun x => C (Epostincr id x ty))).
  rewrite (eval_simple_lvalue_lv H0). apply esc_postincr; auto.
  apply (proj2 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=LV); eauto using simple_leftcontext.
  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H); simpl.
  rewrite <- H5. destruct assert_type_true with (ty:=ty). rewrite H6; simpl.
  rewrite H1; simpl.
  rewrite H2; simpl.
  rewrite (exec_cast_complete H3); simpl.
  rewrite H4; simpl.
  reflexivity.
  reflexivity. trivial. inversion H; auto. reflexivity.

(* Ecomma *)
eapply exec_star_right. apply simple_execs with (C:=(fun x => C (Ecomma x r2 ty))).
  rewrite (eval_simple_rvalue_rv H0). apply esc_comma; auto.
  apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite <- H1. destruct assert_type_true with (ty:=ty). rewrite H2; simpl.
  reflexivity.
  reflexivity. simpl; trivial. inversion H; auto. reflexivity.

(* Eparen *)
eapply exec_star_right. apply simple_execs with (C:=(fun x => C (Eparen x ty))).
  rewrite (eval_simple_rvalue_rv H0). apply esc_paren; auto.
  apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.
  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite <- H1. destruct assert_type_true with (ty:=ty). rewrite H2; simpl.
  reflexivity.
  reflexivity. simpl; trivial. inversion H; auto. reflexivity.

(* Ecall *)
destruct (eval_simple_list_decompose H2) as [vl [Hl Hc]].
eapply exec_star_right.
  eapply exec_star_trans.
    apply simple_execs with (C:=(fun x => C (Ecall x rargs ty))).
    rewrite (eval_simple_rvalue_rv H1). apply esc_call_left; auto.
    eapply eval_simple_list_is_simplelist. apply H2.
    apply (proj1 (simple_steps ge f k e m)) with (C:=fun x:expr => x) (ck:=RV); eauto using simple_leftcontext.

    apply simple_args_steps; auto.
    apply Hl.
    reflexivity.

  simpl. rewrite not_Eval. unfold exec_estep. unfold pick_cstrategy.
  rewrite ((proj1 leftcontext_find_effect) _ _ _ H). simpl.
  rewrite pick_no_vargs; simpl.
  rewrite H0; simpl.
  rewrite H3; simpl.
  rewrite H4. destruct (dec_eq_true type_eq (Tfunction targs tres)). rewrite H5.
  rewrite (exec_cast_args_complete Hc). simpl.
Require Import Axioms.
  rewrite (functional_extensionality (fun x => C x) C); reflexivity.
  reflexivity. simpl; trivial. split; auto.
  clear. revert rargs. induction vl; simpl. trivial. destruct rargs; constructor. simpl; trivial. apply IHvl.
  inversion H; trivial. reflexivity.
Qed.

Theorem cstrategy_exec_complete : forall ge s tr s',
  Cstrategy.step ge s tr s' ->
  only_supported_external s ->
  exec_star ge s tr s'.
induction 1.
intro. apply cstrategy_estep_complete. assumption.
intro. eapply es_step. apply sstep_complete; eauto.
apply es_refl. rewrite E0_right. reflexivity.
Qed. 

