(* Based on Driver.ml, the horrible bits are by Brian Campbell. *)


(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

open Printf
open Clflags

(* Location of the compatibility library *)

let stdlib_path = ref(
  try
    Sys.getenv "COMPCERT_LIBRARY"
  with Not_found ->
    Configuration.stdlib_path)

let command cmd =
  if !option_v then begin
    prerr_string "+ "; prerr_string cmd; prerr_endline ""
  end;
  Sys.command cmd

let quote_options opts =
  String.concat " " (List.rev_map Filename.quote opts)

let safe_remove file =
  try Sys.remove file with Sys_error _ -> ()

(* Printing of error messages *)

let print_error oc msg =
  let print_one_error = function
  | Errors.MSG s -> output_string oc (Camlcoq.camlstring_of_coqstring s)
  | Errors.CTX i -> output_string oc (Camlcoq.extern_atom i)
  in
    List.iter print_one_error msg;
    output_char oc '\n'

(* From C to preprocessed C *)

let preprocess ifile ofile =
  let output =
    if ofile = "-" then "" else sprintf "> %s" ofile in
  let cmd =
    sprintf "%s -D__COMPCERT__ %s %s %s %s"
      Configuration.prepro
      (if Configuration.need_stdlib_wrapper
       then sprintf "-I%s" !stdlib_path
       else "")
      (quote_options !prepro_options)
      ifile output in
  if command cmd <> 0 then begin
    if ofile <> "-" then safe_remove ofile;
    eprintf "Error during preprocessing.\n";
    exit 2
  end

(* Hacked up memcpy.  Fortunately extraction reveals the internal 
   representation. TODO: bound checking *)

let memcpy m (dstb, dsto) (srcb, srco) n =
  let old = m.Memory.Mem.mem_contents in
  {m with Memory.Mem.mem_contents =
    fun b o ->
      if b = dstb then
        let o' = BinInt.coq_Zminus o dsto in
        if Zbool.coq_Zge_bool o dsto && Zbool.coq_Zlt_bool o' n then
          old srcb (BinInt.coq_Zplus srco o')
        else old b o
      else old b o}
;;

(* Prevent generation of inline memcpies. *)
C2C.memcpy_inline_threshold := 0l;;

(* Hacked up memcmp.  Note that bytes of pointers always sort above everything
   else! *)

let byte_cmp a b =
  BinInt.coq_Zcompare (Integers.Byte.unsigned a) (Integers.Byte.unsigned b)

let memcmp m (pb, po) (qb, qo) n =
  let rec aux po qo n =
    if n = 0 then Datatypes.Eq else
    match m.Memory.Mem.mem_contents pb po, m.Memory.Mem.mem_contents qb qo with
    | Memdata.Byte a, Memdata.Byte b ->
        begin match byte_cmp a b with
        | Datatypes.Eq -> aux (BinInt.coq_Zsucc po) (BinInt.coq_Zsucc qo) (n-1)
        | x -> x
        end
    | Memdata.Byte _, Memdata.Pointer _ -> Datatypes.Lt
    | Memdata.Pointer _, Memdata.Byte _ -> Datatypes.Gt
    | Memdata.Pointer (ab,ao,ai), Memdata.Pointer (bb,bo,bi) ->
        (match BinInt.coq_Zcompare ab bb with
         | Datatypes.Eq ->
           (match BinInt.coq_Zcompare (Integers.Int.unsigned ao)
                                      (Integers.Int.unsigned bo) with
            | Datatypes.Eq ->
               (match Compare_dec.nat_compare ai bi with
               | Datatypes.Eq ->
                   aux (BinInt.coq_Zsucc po) (BinInt.coq_Zsucc qo) (n-1)
               | x -> x)
            | x -> x)
          | x -> x)
    | _ -> failwith "memcmp on undefined memory"
  in
  match aux po qo (Int32.to_int (Camlcoq.camlint_of_coqint n)) with
  | Datatypes.Lt -> Values.coq_Vmone
  | Datatypes.Eq -> Values.coq_Vzero
  | Datatypes.Gt -> Values.coq_Vone

(* Hacky printf for C programs *)

let char_of_byte b =
  let i = Camlcoq.camlint_of_z (Integers.Byte.unsigned b) in
  Char.chr (Int32.to_int i)

let retrieve_string m (b,o) =
  let get_byte o =
    match m.Memory.Mem.mem_contents b o with
    | Memdata.Byte b -> char_of_byte b
    | _ -> failwith "Non-byte in string"
  in
  let rec fold f o a =
    match get_byte o with
    | '\000' -> a
    | x -> fold f (BinInt.coq_Zsucc o) (f x a)
  in
  let len = fold (fun _ x -> x+1) o 0 in
  let s = String.create len in
  let _ = fold (fun c i -> String.set s i c; i+1) o 0 in
  s

let printf m (fmtb, fmto) args =
  let get_byte fmto =
    match m.Memory.Mem.mem_contents fmtb fmto with
    | Memdata.Byte b -> char_of_byte b
    | _ -> failwith "Non-byte in printf format"
  in
  let rec conv fmto arg =
    match get_byte fmto with
    | '-' | '+' | ' ' | '#' | '0' | 'h' | 'l' | 'j' | 'z' | 't' | 'L'
    | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9' | '*' | '.' -> 
      conv (BinInt.coq_Zsucc fmto) arg (* all ignored! *)
    | 'd' | 'i' ->
        begin match arg with
        | Values.Vint i -> Int32.to_string (Camlcoq.camlint_of_z i), fmto
        | _ -> failwith "Non-integer given to printf where one expected"
        end
    | 'x' | 'X' as x ->
        begin match arg with
        | Values.Vint i ->
            (if x = 'x' then sprintf "%lx" else sprintf "%lX")
              (Camlcoq.camlint_of_z i), fmto
        | _ -> failwith "Non-integer given to printf where one expected"
        end
    | 's' ->
        begin match arg with
        | Values.Vptr (b,o) -> retrieve_string m (b,o), fmto
        | _ -> failwith "Non-pointer given to printf where string expected"
        end
    | _ -> failwith "Unsupported printf conversion"
  in
  let rec aux fmto args count =
    match get_byte fmto with
    | '%' ->
        begin match args with
        | h::t ->
            let (s,o) = conv (BinInt.coq_Zsucc fmto) h in
            print_string s;
            aux (BinInt.coq_Zsucc o) t (count+String.length s)
        | _ -> failwith "Missing argument to printf"
        end
    | '\000' -> count
    | x -> print_char x; aux (BinInt.coq_Zsucc fmto) args (count+1)
  in
  let count = aux fmto args 0 in
  Values.Vint (Integers.Int.repr (Camlcoq.z_of_camlint (Int32.of_int count)))

let map_opt f e =
  match e with
  | None -> None
  | Some x -> Some (f x)

open Csyntax

(* Apply workarounds in reducible parts of an expression *)

let rec try_expr f e =
  match f e with
  | Some e' -> Some e'
  | None ->
      match e with
      | Efield (e1,id,ty) -> map_opt (fun x -> Efield (x,id,ty)) (try_expr f e1)
      | Evalof (e1,ty) -> map_opt (fun x -> Evalof (x,ty)) (try_expr f e1)
      | Ederef (e1,ty) -> map_opt (fun x -> Ederef (x,ty)) (try_expr f e1)
      | Eaddrof (e1,ty) -> map_opt (fun x -> Eaddrof (x,ty)) (try_expr f e1)
      | Eunop (op,e1,ty) -> map_opt (fun x -> Eunop (op,x,ty)) (try_expr f e1)
      | Ebinop (op,e1,e2,ty) ->
          (match try_expr f e1 with
          | None -> map_opt (fun x -> Ebinop (op,e1,x,ty)) (try_expr f e2)
          | Some x -> Some (Ebinop (op,x,e2,ty)))
      | Ecast (e1,ty) -> map_opt (fun x -> Ecast (x,ty)) (try_expr f e1)
      | Econdition (e1,e2,e3,ty) -> map_opt (fun x -> Econdition (x,e2,e3,ty)) (try_expr f e1)
      | Eassign (e1,e2,ty) ->
          (match try_expr f e1 with
          | None -> map_opt (fun x -> Eassign (e1,x,ty)) (try_expr f e2)
          | Some x -> Some (Eassign (x,e2,ty)))
      | Eassignop (op,e1,e2,ty1,ty2) ->
          (match try_expr f e1 with
          | None -> map_opt (fun x -> Eassignop (op,e1,x,ty1,ty2)) (try_expr f e2)
          | Some x -> Some (Eassignop (op,x,e2,ty1,ty2)))
      | Epostincr (op,e1,ty) -> map_opt (fun x -> Epostincr (op,x,ty)) (try_expr f e1)
      | Ecomma (e1,e2,ty) -> map_opt (fun x -> Ecomma (x,e2,ty)) (try_expr f e1)
      | Ecall (e1,es,ty) ->
          (match try_expr f e1 with
          | None -> map_opt (fun x -> Ecall (e1,x,ty)) (try_exprlist f es)
          | Some x -> Some (Ecall (x,es,ty)))
      | Eparen (e1,ty) -> map_opt (fun x -> Eparen (x,ty)) (try_expr f e1)
      | _ -> None
and try_exprlist f es =
  match es with
  | Enil -> None
  | Econs (e, es) ->
      match try_expr f e with
      | None -> map_opt (fun x -> Econs (e, x)) (try_exprlist f es)
      | Some e' -> Some (Econs (e', es))

let reduce_workaround = function
  | Econdition (Eval(v1, ty1), r1, r2, ty) ->
      begin match Cexec.exec_bool_val v1 ty1 with
      | Errors.OK b ->
          let r = if b then r1 else r2 in
          Some (Ecast (r, ty))
      | Errors.Error msg -> print_error stderr msg; exit 2
      end
  | Ecast (Eval(v1, ty1), Tvoid) ->
    Some (Eval(Values.Vundef, Csyntax.Tvoid))
  | _ -> None

let try_workarounds = try_expr reduce_workaround

(* Execute *)

(* Allow main to be something other than int main(void) *)
let option_relaxmain = ref false

(* Whether to print an exit value *)
let option_exitvalue = ref false

(* Apply known workarounds to bugs in the semantics *)
let option_workaround = ref false

(* Which evaluation strategy to use. *)
let option_strategy = ref Cexec.pick_leftmost_innermost

(* Stuckness check. *)
let option_stuckness = ref true

(* The exit value of the last C file will be used as the exit value of the
   whole program. *)
let exitvalue = ref Int32.zero

let exec_c_files inputs =
  Sections.initialize();
  (* Simplification options *)
  let simplifs =
    "b" (* blocks: mandatory *)
  ^ (if !option_fstruct_passing then "s" else "")
  ^ (if !option_fstruct_assign then "S" else "")
  ^ (if !option_fbitfields then "f" else "")
  ^ (if !option_fpacked_structs then "p" else "")
  in
  let cil_c_of_file (sourcename,ifile) =
    (* Parsing and production of a simplified C AST *)
    let ast =
      match Cparser.Parse.preprocessed_file simplifs sourcename ifile with
      | None -> exit 2
      | Some p -> p in
    (* Remove preprocessed file (always a temp file) *)
    safe_remove ifile;
    (* Save C AST if requested *)
    if !option_dparse then begin
      let ofile = Filename.chop_suffix sourcename ".c" ^ ".parsed.c" in
      let oc = open_out ofile in
      Cparser.Cprint.program (Format.formatter_of_out_channel oc) ast;
      close_out oc
    end;
    ast
  in
  (* Combine all of the CIL ASTs together.  FIXME? Ideally we would do some
     renaming to avoid clashes. *)
  let ast = List.concat (List.map cil_c_of_file inputs) in
  (* Conversion to Csyntax *)
  let csyntax =
    match C2C.convertProgram ast with
    | None -> exit 2
    | Some p -> p in
  flush stderr;
  let ge,s =
    match if !option_relaxmain 
          then Cexec.make_initial_state_relaxed csyntax
          else Cexec.make_initial_state csyntax with
    | Errors.OK (Datatypes.Coq_pair (ge,s)) -> (ge,s)
    | Errors.Error msg -> print_error stderr msg; exit 2
  in
  let is_fn s f = Camlcoq.extern_atom f.AST.ef_id = s in
  let is_printf f =
    try (String.sub (Camlcoq.extern_atom f.AST.ef_id) 0 7) = "printf$"
    with _ -> false in
  let is_memcpy f = 
    try
    let name = Camlcoq.extern_atom f.AST.ef_id in
      name = "memcpy" || (String.sub name 0 16) = "__builtin_memcpy"
    with _ -> false
  in
  let rec aux s =
    match Cexec.is_final_state s with
    | None -> begin
        match s with
        (* Very special case for csmith *)
        | Csem.Callstate (Csyntax.External (f,_,_),(Values.Vptr (b,o)::t), k, m)
                                                    when is_printf f ->
            let ret = printf m (b,o) t in
            aux (Csem.Returnstate (ret, k, m))
        (* Rather dodgy case for memcpy *)
        | Csem.Callstate (Csyntax.External (f,_,_),
            [Values.Vptr (dstb,dsto); Values.Vptr (srcb,srco); Values.Vint n],
            k, m) when is_memcpy f ->
            let m' = memcpy m (dstb,dsto) (srcb,srco) n in
            aux (Csem.Returnstate (Values.Vptr (dstb,dsto), k, m'))
        | Csem.Callstate (Csyntax.External (f,_,_),
            [Values.Vptr (ab,ao); Values.Vptr (bb,bo); Values.Vint n],
            k, m) when is_fn "memcmp" f ->
            let r = memcmp m (ab,ao) (bb,bo) n in
            aux (Csem.Returnstate (r, k, m))
        (* abort and exit for misc gcc tests *)
        | Csem.Callstate (Csyntax.External (f,_,_), _, k, m)
                                                    when is_fn "abort" f ->
            Printf.eprintf "Aborted.\n"; exit 2
        | Csem.Callstate (Csyntax.External (f,_,_), [Values.Vint x], k, m)
                                                    when is_fn "exit" f ->
            x
        (* Work around bugs in the semantics. *)
        | Csem.ExprState (f,r,k,e,m) when !option_workaround ->
            (match try_workarounds r with
            | Some r' -> aux (Csem.ExprState (f,r',k,e,m))
            | None -> real_step s)
        (* Otherwise, actually execute a step of the real semantics *)
        | _ -> real_step s
        end
    | Some r -> r
  and real_step s =
    match Cexec.exec_step ge !option_strategy s !option_stuckness with
    | Errors.OK (Datatypes.Coq_pair (tr,s')) -> aux s'
    | Errors.Error msg -> print_error stderr msg; exit 2
  in
  let res = aux s in
  exitvalue := Camlcoq.camlint_of_coqint res;
  if !option_exitvalue then
    Printf.printf "Execution completed: %ld\n" !exitvalue


(* Processing of a .c file *)

let inputs = ref [ ]

let process_c_file sourcename =
  let prefixname = Filename.chop_suffix sourcename ".c" in
  if !option_E then begin
    preprocess sourcename "-"
  end else begin
    let preproname = Filename.temp_file "compcert" ".i" in
    preprocess sourcename preproname;
    inputs := (sourcename, preproname)::!inputs
  end

(* Command-line parsing *)

let usage_string =
"cexec [options] <source files>
Recognized source files:
  .c             C source file
Processing options:
  -E             Preprocess only, send result to standard output
Preprocessing options:
  -I<dir>        Add <dir> to search path for #include files
  -D<symb>=<val> Define preprocessor symbol
  -U<symb>       Undefine preprocessor symbol
Language support options (use -fno-<opt> to turn off -f<opt>) :
  -fbitfields    Emulate bit fields in structs [off]
  -flonglong     Partial emulation of 'long long' types [on]
  -fstruct-passing  Emulate passing structs and unions by value [off]
  -fstruct-assign   Emulate assignment between structs or unions [off]
  -fvararg-calls Emulate calls to variable-argument functions [on]
  -fpacked-structs  Emulate packed structs [off]
Code generation options:
  -fmadd         Use fused multiply-add and multiply-sub instructions [off]
  -fsmall-data <n>  Set maximal size <n> for allocation in small data area
  -fsmall-const <n>  Set maximal size <n> for allocation in small constant area
Tracing options:
  -dparse        Save C file after parsing and elaboration in <file>.parse.c
  -dc            Save generated Compcert C in <file>.compcert.c
Debug options:
  -drelaxmain    Don't type check main function
  -dexitvalue    Print exit value of program
  -dworkaround   Partial workaround known bugs with ?: and (void)
General options:
  -stdlib <dir>  Set the path of the Compcert run-time library [MacOS X only]
  -v             Print external commands before invoking them

If multiple .c files are given they will be concatenated after conversion to
CIL ASTs.  This may not have the desired effect where there are name clashes.
"

type action =
  | Set of bool ref
  | Unset of bool ref
  | Self of (string -> unit)
  | String of (string -> unit)
  | Integer of (int -> unit)

let rec find_action s = function
  | [] -> None
  | (re, act) :: rem ->
      if Str.string_match re s 0 then Some act else find_action s rem

let parse_cmdline spec usage =
  let acts = List.map (fun (pat, act) -> (Str.regexp pat, act)) spec in
  let error () =
    eprintf "Usage: %s" usage;
    exit 2 in
  let rec parse i =
    if i < Array.length Sys.argv then begin
      let s = Sys.argv.(i) in
      match find_action s acts with
      | None ->
          if s <> "-help" && s <> "--help" 
          then eprintf "Unknown argument `%s'\n" s;
          error ()
      | Some(Set r) ->
          r := true; parse (i+1)
      | Some(Unset r) ->
          r := false; parse (i+1)
      | Some(Self fn) ->
          fn s; parse (i+1)
      | Some(String fn) ->
          if i + 1 < Array.length Sys.argv then begin
            fn Sys.argv.(i+1); parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
      | Some(Integer fn) ->
          if i + 1 < Array.length Sys.argv then begin
            let n =
              try
                int_of_string Sys.argv.(i+1)
              with Failure _ ->
                eprintf "Argument to option `%s' must be an integer\n" s;
                error()
            in
            fn n; parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
    end
  in parse 1

let cmdline_actions =
  let f_opt name ref =
    ["-f" ^ name ^ "$", Set ref; "-fno-" ^ name ^ "$", Unset ref] in
  [
  "-I$", String(fun s -> prepro_options := s :: "-I" :: !prepro_options);
  "-D$", String(fun s -> prepro_options := s :: "-D" :: !prepro_options);
  "-U$", String(fun s -> prepro_options := s :: "-U" :: !prepro_options);
  "-nostdinc$", Self(fun s -> prepro_options := s :: !prepro_options);
  "-[IDU].", Self(fun s -> prepro_options := s :: !prepro_options);
  "-[lL].", Self(fun s -> linker_options := s :: !linker_options);
  "-o$", String(fun s -> exe_name := s);
  "-stdlib$", String(fun s -> stdlib_path := s);
  "-dparse$", Set option_dparse;
  "-dc$", Set option_dcmedium;
  "-dclight$", Set option_dclight;
  "-dcminor", Set option_dcminor;
  "-drtl$", Set option_drtl;
  "-dtailcall$", Set option_dtailcall;
  "-dcastopt$", Set option_dcastopt;
  "-dconstprop$", Set option_dconstprop;
  "-dcse$", Set option_dcse;
  "-dalloc$", Set option_dalloc;
  "-dmach$", Set option_dmach;
  "-dasm$", Set option_dasm;
  "-dexitvalue$", Set option_exitvalue;
  "-drelaxmain$", Set option_relaxmain;
  "-dworkaround$", Set option_workaround;
  "-E$", Set option_E;
  "-S$", Set option_S;
  "-c$", Set option_c;
  "-v$", Set option_v;
  ".*\\.c$", Self (fun s -> process_c_file s);
  "-fsmall-data$", Integer(fun n -> option_small_data := n);
  "-fsmall-const$", Integer(fun n -> option_small_const := n);
  "-fcstrategy$", Self(fun _ -> option_strategy := Cexec.pick_cstrategy;
                                option_stuckness := false)
  ]
  @ f_opt "longlong" option_flonglong
  @ f_opt "struct-passing" option_fstruct_passing
  @ f_opt "struct-assign" option_fstruct_assign
  @ f_opt "bitfields" option_fbitfields
  @ f_opt "vararg-calls" option_fvararg_calls
  @ f_opt "madd" option_fmadd
  @ f_opt "packed-structs" option_fpacked_structs

let _ =
  Gc.set { (Gc.get()) with Gc.minor_heap_size = 524288 };
  Cparser.Machine.config :=
    begin match Configuration.arch with
    | "powerpc" -> Cparser.Machine.ppc_32_bigendian
    | "arm"     -> Cparser.Machine.arm_littleendian
    | "ia32"    -> Cparser.Machine.x86_32
    | _         -> assert false
    end;
  Cparser.Builtins.set C2C.builtins;
  CPragmas.initialize();
  parse_cmdline cmdline_actions usage_string;
  match !inputs with
  | [ ] -> (prerr_endline "No C files!"; exit 2)
  | x -> exec_c_files (List.rev x); exit (Int32.to_int !exitvalue)
